(****************************************************************************)
(*                                                                          *)
(* Ce fichier contient les fonctions ncessaires au calcule des BoundingBox *)
(*                                                                          *)
(* Un BoundingBox se prsente sous la forme d'un 6-uplet de flottants.      *)
(* Les 3 premires coordonnes sont celles du coin infrieur du BoundingBox *)
(* dans le repre (X, Y, Z).                                                *)
(* Les 3 dernires sont celles du coin suprieur.                           *)
(*                                                                          *)
(****************************************************************************)



open Types
open List
open Triplet
open Parametres



(********************************************************)
(* Calcule la somme des lments flottants d'un tableau *)
(********************************************************)
(* Type : float array -> float                          *)
(********************************************************)
let sumTab = Array.fold_left (fun v s -> v +. s) 0.0



(****************************************************)
(* Calcule le minimum ou le maximum de 4 valeurs    *)
(****************************************************)
(* Type : float -> float -> float -> float -> float *)
(****************************************************)
let min_of_4 x1 x2 x3 x4 = min (min x1 x2) (min x3 x4)
let max_of_4 x1 x2 x3 x4 = max (max x1 x2) (max x3 x4)



(***********************************************************)
(* Calcule le BoundingBox de 2 objets                      *)
(***********************************************************)
(* Type : float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float    *)
(***********************************************************)
let bounding_fusion (minX1, minY1, minZ1, maxX1, maxY1, maxZ1) (minX2, minY2, minZ2, maxX2, maxY2, maxZ2) = (min minX1 minX2,
													     min minY1 minY2,
													     min minZ1 minZ2,
													     max maxX1 maxX2,
													     max maxY1 maxY2,
													     max maxZ1 maxZ2)



(********************************************************)
(* Calcule le BoundingBox d'une liste de points         *)
(********************************************************)
(* Type : point list ->                                 *)
(*        float * float * float * float * float * float *)
(********************************************************)
let boundingPolyline lst = match lst
			   with
			     | []     -> (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
			     | hd::tl -> let rec extremPolyline lst bounding = match lst
									       with
										 | []     -> bounding
										 | hd::tl -> extremPolyline tl
											     (bounding_fusion bounding (hd#getX,
															hd#getY,
															hd#getZ,
															hd#getX,
															hd#getY,
															hd#getZ)
											     )
					 in
					   extremPolyline tl (hd#getX, hd#getY, hd#getZ, hd#getX, hd#getY, hd#getZ) 



(***********************************************************************)
(* Met  jour la valeur de l ou t.(i) par le calcul de max(l, max-min) *)
(* Utilises pour le calcul des BB des Grids                           *)
(***********************************************************************)
(* Type : float ref -> float -> float -> unit                          *)
(***********************************************************************)
let max_diff_update_ref l ma mi = l := (max !l (ma -. mi))
(***********************************************************************)
(* Type : float array -> int -> float -> float -> unit                 *)
(***********************************************************************)
let max_diff_update_tab t i ma mi = t.(i) <- (max t.(i) (ma -. mi))



(*************************************************************************)
(* Permute les coordonnes du boundingBox pour passer du repre          *)
(* (X, Y, Z) au repre (axis1, axis2, ...)                               *)
(*************************************************************************)
(* Type : float * float * float * float * float * float -> axe -> axe -> *)
(*        float * float * float * float * float * float                  *)
(*************************************************************************)
let desordre (x1, y1, z1, x2, y2, z2) axis1 axis2 = match (axis1, axis2)
						    with
						      | (AxeX, AxeY) -> (x1, y1, z1, x2, y2, z2)
						      | (AxeX, AxeZ) -> (x1, z1, y1, x2, z2, y2)
						      | (AxeY, AxeX) -> (y1, x1, z1, y2, x2, z2)
						      | (AxeY, AxeZ) -> (y1, z1, x1, y2, z2, x2)
						      | (AxeZ, AxeX) -> (z1, x1, y1, z2, x2, y2)
						      | (AxeZ, AxeY) -> (z1, y1, x1, z2, y2, x2)
						      | _            -> raise (Erreur "BoundingBox : repre invalide")



(*************************************************************************)
(* Permute les coordonnes du boundingBox pour passer du repre          *)
(* (axis1, axis2, ...) au repre (X, Y, Z)                               *)
(*************************************************************************)
(* Type : float * float * float * float * float * float -> axe -> axe -> *)
(*        float * float * float * float * float * float                  *)
(*************************************************************************)
let reordre (x1, y1, z1, x2, y2, z2) axis1 axis2 = match (axis1, axis2)
						    with
						      | (AxeX, AxeY) -> (x1, y1, z1, x2, y2, z2)
						      | (AxeX, AxeZ) -> (x1, z1, y1, x2, z2, y2)
						      | (AxeY, AxeX) -> (y1, x1, z1, y2, x2, z2)
						      | (AxeY, AxeZ) -> (z1, x1, y1, z2, x2, y2)
						      | (AxeZ, AxeX) -> (y1, z1, x1, y2, z2, x2)
						      | (AxeZ, AxeY) -> (z1, y1, x1, z2, y2, x2)
						      | _            -> raise (Erreur "BoundingBox : repre invalide")



(***********************************************************)
(* Dtermine l'axe axis2 tel que (axis, axis2) soit direct *)
(***********************************************************)
(* Type : axe -> axe                                       *)
(***********************************************************)
let axe_suiv axis = match axis
		    with
		      | AxeX    -> AxeY
		      | AxeY    -> AxeZ
		      | AxeZ    -> AxeX
		      | AxeNone -> AxeNone



(**************************************************************)
(* Permute les coordonnes pour passer du repre              *)
(* (axis1, axis2, ...) au repre (X, Y, Z)                    *)
(* Ralise en quelque sorte l'inverse de la fonction desordre *)
(**************************************************************)
(* Type : float * float * float -> axe -> axe ->              *)
(*        float * float *float                                *)
(**************************************************************)
let permut (x, y, z) axis1 axis2 = match (axis1, axis2)
				   with
				     | (AxeX, AxeY) -> (x, y, z)
				     | (AxeX, AxeZ) -> (x, z, y)
				     | (AxeY, AxeX) -> (y, x, z)
				     | (AxeY, AxeZ) -> (z, x, y)
				     | (AxeZ, AxeX) -> (y, z, x)
				     | (AxeZ, AxeY) -> (z, y, x)
				     | (_, _)       -> raise (Erreur "BoundingBox : repre invalide")



(************************************************************)
(* Calcule la 1re partie de la translation sur un axe pour *)
(* passer d'un lment  un autre autre dans un Grid.       *)
(* align est l'alignement de l'objet sur l'axe, min et max  *)
(* les coordonnes sur cet axe du boundingbox de l'objet    *)
(************************************************************)
(* Type : align -> float -> float -> float                  *)
(************************************************************)
let sbt algn min max = (match algn
                        with
			  | Left   -> (max -. min) /. 2.0
			  | Middle -> 0.0
			  | Right  -> (min -. max) /. 2.0
		       ) -. ((min +. max) /. 2.0)



(************************************************************)
(* Calcule la 2nde partie de la translation sur un axe pour *)
(* passer d'un lment  un autre autre dans un Grid.       *)
(* align est l'alignement de l'objet sur l'axe, nouv et anc *)
(* les tailles sur cet axe du boundingbox de l'objet et de  *)
(* son prdcesseur.                                        *)
(************************************************************)
(* Type : align -> float -> float -> float                  *)
(************************************************************)
let sbt2 algn anc nouv = match algn
			 with
			   | Left   -> (anc -. nouv) /. 2.0
			   | Middle -> 0.0
			   | Right  -> (nouv -. anc) /. 2.0



(*************************************************************)
(* Calcule le boundingBox d'un Grid dans la base (X, Y, Z)   *)
(*  partir du centre du Grid center et des dimensions dim1, *)
(* dim2 et dim3 dans la base (axis1, axis2, ...)             *)
(*************************************************************)
(* Type : point -> float -> float -> float -> axe -> axe ->  *)
(*        float * float * float * float * float * float      *)
(*************************************************************)
let boundingBox_of center dim1 dim2 dim3 axis1 axis2 = let dims = permut (dim1 /. 2.0, dim2 /. 2.0, dim3 /. 2.0) axis1 axis2
						       in 
						       let (minX, minY, minZ) = diff_triplet center#getAll dims
						       and (maxX, maxY, maxZ) =  add_triplet center#getAll dims
						       in
							 (minX, minY, minZ, maxX, maxY, maxZ)



(***********************************************************)
(* Intervertit les valeurs min et max si min > max         *)
(***********************************************************)
(* Type : float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float    *)
(***********************************************************)
let bonOrdre (minX, minY, minZ, maxX, maxY, maxZ) = (min minX maxX, min minY maxY, min minZ maxZ, max minX maxX, max minY maxY, max minZ maxZ)



(***********************************************************)
(* Dtermine le centre d'un boundingBox.                   *)
(***********************************************************)
(* Type : float * float * float * float * float * float -> *)
(*        float * float * float                            *)
(***********************************************************)
let middle (minX, minY, minZ, maxX, maxY, maxZ) = ((maxX +. minX) /. 2.0,
						   (maxY +. minY) /. 2.0,
						   (maxZ +. minZ) /. 2.0)



(*******************************************************************)
(* Fonction rcurrente pour calculer les BoundingBox de listes     *)
(* d'objets lst et d'un objet ayant le boundingbox bounding        *)
(*******************************************************************)
(* Type : list -> float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float            *)
(*******************************************************************)
let rec _boundingList lst bounding = match lst
				      with
					| []   -> bounding
					| t::q -> let bounding2 = boundingBox t
						  in
						    _boundingList q (bounding_fusion bounding bounding2)



(*******************************************************************)
(* Calcule le BoundingBox d'une liste d'objets lst                 *)
(*******************************************************************)
(* Type : list -> float * float * float * float * float * float -> *)
(*        float * float * float * float * float * float            *)
(*******************************************************************)
and  boundingList lst = match lst
			with
			  | []     -> (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
			  | hd::tl -> _boundingList tl (boundingBox hd)



(********************************************************)
(* Calcule le BoundingBox d'un objet obj                *)
(********************************************************)
(* Type : objet ->                                      *)
(*        float * float * float * float * float * float *)
(********************************************************)
and boundingBox obj =
  match obj with
    | Vide -> (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)

    | Primitive (prim, _, _, cont) -> (* On grandit un peu le rayon de cnes, sphres, ... pour pas que le contenu ne dborde au niveau des facettes *)
	let coeffRayonCont cotes = cos (Parametres.pi /. (float_of_int cotes))
	in
	let (minX, minY, minZ, maxX, maxY, maxZ), zoom, trans = match cont
	with
	  | None           -> (0.0, 0.0, 0.0, 0.0, 0.0, 0.0), ref nan, new point nan nan nan
	  | Some (c, h, l) -> let bb = boundingList l
	    in
	      c#setAll (Triplet.oppose_triplet (middle bb));
	      bb, h, c
	in
	let (midX, midY, midZ) = middle (minX, minY, minZ, maxX, maxY, maxZ)
	in
	  (match prim
	   with
	     | Box pt                  -> if cont = None
	       then
		 begin
		   if pt#isDefault
		   then
		     pt#setAll (1.0, 1.0, 1.0)
		 end
	       else
		 begin
		   if pt#isDefault
		   then
		     pt#setAll ((maxX -. minX) /. 2.0, (maxY -. minY) /. 2.0, (maxZ -. minZ) /. 2.0)
		   else
		     begin
		       let minR = min (min (abs_float ((2.0 *. pt#getX) /. (maxX -. minX)))
					 (abs_float ((2.0 *. pt#getY) /. (maxY -. minY))))
				    (abs_float ((2.0 *. pt#getZ) /. (maxZ -. minZ)))
		       in
			 if minR < 1.0
			 then
			   zoom := minR
		     end
		 end;
		 
		 (-.pt#getX, -.pt#getY, -.pt#getZ,
		  pt#getX,   pt#getY,   pt#getZ)
		 
	     | Cone (r, h, _, s)       -> if cont = None
	       then
		 begin
		   if !r == nan
		   then
		     r := 1.0;
		   if !h == nan
		   then
		     h := 1.0
		 end
	       else
		 begin
		   let coeff = coeffRayonCont !s
		   and diag  = 0.5 *. (sqrt (let dX, dZ = maxX -. minX, maxZ -. minZ
					     in
					       dX*.dX +. dZ*.dZ))
		   and haut  = 0.5 *. (maxY -. minY)
		   in
		   let zoomOfR_H r h = min 1.0 (1.0 /. ((haut/.h) +. (diag/.(r*.coeff))))
		   in
		     if !r == nan
		     then
		       begin
			 if !h == nan
			 then
			   begin
			     r := (diag /. coeff) +. haut;
			     h := !r
			   end
			 else
			   begin
			     if !h > haut
			     then
			       r := diag /. (coeff *. (1.0 -. (haut /. !h)))
			     else
			       begin
				 r    := !h;
				 zoom := zoomOfR_H !r !h
			       end
			   end
		       end
		     else
		       begin
			 if !h == nan
			 then
			   begin
			     if !r > (diag /. coeff)
			     then
			       h := haut /. (1.0 -. (diag /. (!r *. coeff)))
			     else
			       begin
				 h    := !r;
				 zoom := zoomOfR_H !r !h
			       end
			   end
			 else
			   zoom := zoomOfR_H !r !h
		       end;
		     
		     trans#updateY (-. !h *. diag /. (!r*.coeff))
		 end;
		 
		 (-. !r, -. !h, -. !r,
		  !r,    !h,    !r)
		 
	     | Cylinder (r, h, _, s)   -> if cont = None
	       then
		 begin
		   if !r == nan
		   then
		     r := 1.0;
		   if !h == nan
		   then
		     h := 1.0
		 end
	       else
		 begin
		   let coeff = coeffRayonCont !s
		   and diag  = 0.5 *. (sqrt (let dX = maxX -. minX
					     and dZ = maxZ -. minZ
					     in
					       dX*.dX +. dZ*.dZ))
		   in
		   let defRayon = diag /. coeff
		   and defHaut  = 0.5 *. (maxY -. minY)
		   in
		   let rR = (abs_float (!r /. defRayon))
		   and rH = (abs_float (!h /. defHaut))
		   in
		     if !r == nan
		     then
		       begin
			 if !h == nan
			 then
			   begin
			     r := defRayon;
			     h := defHaut
			   end
			 else
			   begin
			     if rH < 1.0
			     then
			       zoom := rH;
			     r := defRayon *. !zoom
			   end
		       end
		     else
		       begin
			 if !h == nan
			 then
			   begin
			     if rR < 1.0
			     then
			       zoom := rR;
			     h := defHaut *. !zoom
			   end
			 else
			   begin
			     let minR = min rR rH
			     in
			       if minR < 1.0
			       then
				 zoom := minR
			   end
		       end
		 end;
		 
		 (-. !r, -. !h, -. !r,
		  !r,    !h,    !r)
		 
	     | Disc (r, _)             -> (-. !r, 0.0, -. !r, !r, 0.0, !r)
		 
	     | String _                -> ( 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
		 
	     | Empty (pt, _)           -> if cont = None
	       then
		 begin
		   if pt#isDefault
		   then
		     pt#setAll (1.0, 1.0, 1.0)
		 end
	       else
		 begin
		   if pt#isDefault
		   then
		     pt#setAll ((maxX -. minX) /. 2.0, (maxY -. minY) /. 2.0, (maxZ -. minZ) /. 2.0)
		   else
		     begin
		       let minR = min (min (abs_float ((2.0 *. pt#getX) /. (maxX -. minX)))
					 (abs_float ((2.0 *. pt#getY) /. (maxY -. minY))))
				    (abs_float ((2.0 *. pt#getZ) /. (maxZ -. minZ)))
		       in
			 if minR < 1.0
			 then
			   zoom := minR
		     end
		 end;
		 
		 (-.pt#getX, -.pt#getY, -.pt#getZ,
		  pt#getX,   pt#getY,   pt#getZ)
		 
	     | Frustum (r, h, t, _, s) -> if !t == nan
	       then
		 t := 0.25;
		 if cont = None
		 then
		   begin
		     if !r == nan
		     then
		       r := 1.0;
		     if !h == nan
		     then
		       h := 1.0
		   end
		 else
		   begin
		     let coeff = coeffRayonCont !s
		     and diag  = 0.5 *. (sqrt (let dX, dZ = maxX -. minX, maxZ -. minZ
					       in
						 dX*.dX +. dZ*.dZ))
		     and haut  = 0.5 *. (maxY -. minY)
		     in
		       if !r == nan
		       then
			 begin
			   if !h == nan
			   then
			     begin
			       h := haut;
			       r := diag /. (coeff *. !t)
			     end
			   else
			     begin
			       if !h >= haut
			       then
				 begin
				   r := 2.*.diag/.(coeff*.(2.+.(!t-.1.)*.(1. +. haut /. !h)))
				 end
			       else
				 begin
				   zoom := !h /. haut;
				   r    := diag /. (!t *. coeff)
				 end
			     end
			 end
		       else
			 begin
			   if !h == nan
			   then
			     begin
			       if !r > (diag /. coeff)
			       then
				 begin
				   h := max haut (haut *. (1. -. !t) /. (1. -. (diag/.(!r*.coeff))));
				   trans#updateY (-. !h +. haut)
				 end
			       else
				 begin
				   zoom := !t *. !r *. coeff /. diag;
				   h    := !zoom *. haut
				 end
			     end
			   else
			     begin
			       if haut > !h
			       then
				 zoom := !h /. haut;
			       if !zoom*.diag > !r *. coeff *. (1. -. !zoom *. haut *.(1. -. !t) /. !h)
			       then
				 zoom := !h /. (!h*.diag/.(!r*.coeff) +. haut*.(1.-. !t));
			       trans#updateY (max 0. ((((2.*. !h /. (!t-.1.)) *. (!zoom*.diag/.(!r*.coeff) -. 1.) -. !h)/. !zoom) -. haut))
			     end
			 end
		   end;
		 
		 (-. !r, -. !h, -. !r,
		  !r,    !h,    !r)
		 
	     | Polygone (l, _)         -> boundingPolyline !l
		 
	     | Polyline l              -> boundingPolyline !l
		 
	     | Sphere (r, sl, st)      -> if cont = None
	       then
		 begin
		   if !r == nan
		   then
		     r := 1.0
		 end
	       else
		 begin
		   let coeff = (coeffRayonCont !sl) *. (coeffRayonCont !st)
		   and diag  = 0.5 *. (sqrt (let dX = maxX -. minX
					     and dY = maxY -. minY
					     and dZ = maxZ -. minZ
					     in
					       dX*.dX +. dY*.dY +. dZ*.dZ))
		   in
		   let defRayon = diag /. coeff
		   in
		     if !r == nan
		     then
		       r := defRayon
		     else
		       begin
			 let rR = (abs_float (!r /. defRayon))
			 in
			   if rR < 1.0
			   then
			     zoom := rR
		       end
		 end;
		 
		 (-. !r, -. !r, -. !r,
		  !r,    !r,    !r)
	  )
	  
    | Transformation (trans, objTrans) -> 
	(match trans
	 with
	     (* Rotation d'angle theta, d'axe axis et de centre O *)
	   | Rotation (axis, theta) -> let (min1, min2, min3, max1, max2, max3) = desordre (boundingBox objTrans) axis (axe_suiv axis)
				       and theta_radian = pi *. theta /. 180.0
	     in
	     let cos_theta = cos theta_radian
	     and sin_theta = sin theta_radian
	     in
	     let rot = function x -> function y -> (cos_theta *. x -. sin_theta *. y, sin_theta *. x +. cos_theta *. y)
	     in
	     let (y1, z1) = rot min2 min3
	     and (y2, z2) = rot max2 min3
	     and (y3, z3) = rot max2 max3
	     and (y4, z4) = rot min2 max3
	     in
	       reordre (min1, min_of_4 y1 y2 y3 y4, min_of_4 z1 z2 z3 z4, max1, max_of_4 y1 y2 y3 y4, max_of_4 z1 z2 z3 z4) 
		 axis
		 (axe_suiv axis)
		 
	   (* Homothtie de centre O et de rapport pt#getX sur l'axe X etc... *)
	   | Scale pt -> let (minX, minY, minZ, maxX, maxY, maxZ) = boundingBox objTrans
	     in
	       bonOrdre (minX *. pt#getX, minY *. pt#getY, minZ *. pt#getZ,
			 maxX *. pt#getX, maxY *. pt#getY, maxZ *. pt#getZ)
		 
	   (* Translation de vecteur pt *)
	   | Translation pt -> let (minX, minY, minZ, maxX, maxY, maxZ) = boundingBox objTrans
	     in
	       (minX +. pt#getX, minY +. pt#getY, minZ +. pt#getZ,
		maxX +. pt#getX, maxY +. pt#getY, maxZ +. pt#getZ)
	)
	
    | Groupe (_, objGrp) -> boundingList objGrp
	
    | Grid (pt, grid, coul, objList) ->
	(match grid
	 with
	     
	   | Grid0 (align1, align2, align3) ->
	       let col    = ref 0.0 
	       and row    = ref 0.0 
	       and stage  = ref 0.0
	       in 
	       let (x_pred, y_pred, z_pred) = (ref 0.0, ref 0.0, ref 0.0)
	       in
		 (* Calcule de la taille du Grid0 et de la 1re partie de la translation pour passer d'un objet au suivant *)
		 List.iter (function (obj, vect) -> let (min1, min2, min3, max1, max2, max3) = boundingBox obj
			    in
			      max_diff_update_ref col   max1 min1;
			      max_diff_update_ref row   max2 min2;
			      max_diff_update_ref stage max3 min3;
			      let (x, y, z) = (sbt align1 min1 max1,
					       sbt align2 min2 max2,
					       sbt align3 min3 max3)
			      in
				vect#setAll (x -. !x_pred,
					     y -. !y_pred,
					     z -. !z_pred);
				x_pred := x;
				y_pred := y;
				z_pred := z;
			   ) objList;
		 
		 (* Calcule 2nde partie de la translation pour passer d'un objet au suivant *)
		 (match objList
		  with
		    | []           -> ()
		    | (_, vect)::_ -> vect#updateAll (sbt2 align1 0.0   !col,
						      sbt2 align2 0.0   !row,
						      sbt2 align3 0.0 !stage);
		 );
		 boundingBox_of pt !col !row !stage AxeX AxeY
		   
	   | Grid1 (axis1, align2, align3, space1) ->
	       let dim   = List.length objList
	       and axis2 = axe_suiv axis1
	       in
	       let col    = Array.make dim 0.0 
	       and row    = ref 0.0 
	       and stage  = ref 0.0
	       and indice = ref 0 
	       in 
	       let (x_pred, y_pred, z_pred) = (ref 0.0, ref 0.0, ref 0.0)
	       in
		 (* Calcule de la taille de chaque case du  Grid1 et de la 1re partie de la translation pour passer d'un objet au suivant *)
		 List.iter (function (obj, vect) -> let (min1, min2, min3, max1, max2, max3) = desordre (boundingBox obj) axis1 axis2
			    in
			      max_diff_update_tab col !indice max1 min1;
			      max_diff_update_ref row         max2 min2;
			      max_diff_update_ref stage       max3 min3;
			      let (x, y, z) = (sbt Middle min1 max1,
					       sbt align2 min2 max2,
					       sbt align3 min3 max3)
			      in
				vect#setAll (permut (x -. !x_pred,
						     y -. !y_pred,
						     z -. !z_pred)
					       axis1
					       axis2);
				x_pred := x;
				y_pred := y;
				z_pred := z;
 				incr indice
			   ) objList;
		 
		 (* Calcule 2nde partie de la translation pour passer d'un objet au suivant *)
		 let sumCol   = (sumTab col)   +. (float_of_int (dim-1)) *. space1
		 in
		   (match objList
		    with
		      | []            -> ()
		      | (_, vect)::tl -> vect#updateAll (permut ((col.(0)   -.   sumCol) /. 2.0,
								 sbt2 align2 0.0   !row,
								 sbt2 align3 0.0 !stage)
							   axis1
							   axis2);
			  indice := 1;
			  List.iter (function (_, vect) -> let delta = ((col.(!indice-1) +. col.(!indice)) /. 2.0 +. space1,
									0.0,
									0.0)
				     in
				       vect#updateAll (permut delta axis1 axis2);
				       incr indice
				    ) tl
		   );
		   boundingBox_of pt sumCol !row !stage axis1 axis2
	   | Grid2 (axis1, axis2, align1, align2, align3, space1, space2, ratio1, ratio2) ->
	       let col    = Array.make ratio1 0.0 
	       and row    = Array.make ratio2 0.0 
	       and stage  = ref 0.0
	       and indice = ref 0 
	       in 
	       let (x_pred, y_pred, z_pred) = (ref 0.0, ref 0.0, ref 0.0)
	       in
		 (* Calcule de la taille de chaque case du  Grid2 et de la 1re partie de la translation pour passer d'un objet au suivant *)
		 List.iter (function (obj, vect) -> let (min1, min2, min3, max1, max2, max3) = desordre (boundingBox obj) axis1 axis2
						    and i = !indice mod ratio1
						    and j = !indice / ratio1
			    in
			      max_diff_update_tab col i max1 min1;
			      max_diff_update_tab row j max2 min2;
			      max_diff_update_ref stage max3 min3;
			      let (x, y, z) = (sbt align1 min1 max1,
					       sbt align2 min2 max2,
					       sbt align3 min3 max3)
			      in
				vect#setAll (permut (x -. !x_pred,
						     y -. !y_pred,
						     z -. !z_pred)
					       axis1
					       axis2);
				x_pred := x;
				y_pred := y;
				z_pred := z;
 				incr indice
			   ) objList;
		 
		 (* Calcule 2nde partie de la translation pour passer d'un objet au suivant *)
		 let sumCol   = (sumTab col) +. (float_of_int (ratio1-1)) *. space1
		 and sumRow   = (sumTab row) +. (float_of_int (ratio2-1)) *. space2
		 in
		   (match objList
		    with
		      | []            -> ()
		      | (_, vect)::tl -> vect#updateAll (permut ((col.(0) -. sumCol) /. 2.0 +. (sbt2 align1 0.0 col.(0)),
								 (row.(0) -. sumRow) /. 2.0 +. (sbt2 align2 0.0 row.(0)),
								 sbt2 align3 0.0 !stage)
							   axis1
							   axis2);
			  indice := 1;
			  List.iter (function (_, vect) -> let i     =  !indice     mod ratio1 
							   and i_old = (!indice -1) mod ratio1 
							   and j     =  !indice      / ratio1
							   and j_old = (!indice - 1) / ratio1
				     in
				     let delta = (if i = 0
						  then ((col.(ratio1-1) +. col.(0)) /. 2.0 -. sumCol,
							(row.(j-1) +. row.(j)) /. 2.0 +. space2,
							0.0)
						  else ((col.(i-1) +. col.(i)) /. 2.0 +. space1,
							0.0,
							0.0)
						 )
				     and delta2 = (sbt2 align1   col.(i_old) col.(i),
						   sbt2 align2   row.(j_old) row.(j),
						   0.0)
				     in
				       vect#updateAll (permut (add_triplet delta delta2)
							 axis1
							 axis2);
				       incr indice
				    ) tl
		   );
		   boundingBox_of pt sumCol sumRow !stage axis1 axis2
		     
	   | Grid3 (axis1, axis2, align1, align2, align3, space1, space2, space3, ratio1, ratio2, ratio3) ->
	       let col    = Array.make ratio1 0.0 
	       and row    = Array.make ratio2 0.0 
	       and stage  = Array.make ratio3 0.0
	       and indice = ref 0 
	       in 
	       let (x_pred, y_pred, z_pred) = (ref 0.0, ref 0.0, ref 0.0)
	       in
		 (* Calcule de la taille de chaque case du  Grid3 et de la 1re partie de la translation pour passer d'un objet au suivant *)
		 List.iter (function (obj, vect) -> let (min1, min2, min3, max1, max2, max3) = desordre (boundingBox obj) axis1 axis2
						    and i = !indice mod ratio1
						    and j = (!indice / ratio1) mod ratio2
						    and k = !indice / (ratio1 * ratio2) 
			    in
			      max_diff_update_tab col   i max1 min1;
			      max_diff_update_tab row   j max2 min2;
			      max_diff_update_tab stage k max3 min3;
			      let (x, y, z) = (sbt align1 min1 max1,
					       sbt align2 min2 max2,
					       sbt align3 min3 max3)
			      in
				vect#updateAll (permut (x -. !x_pred,
							y -. !y_pred,
							z -. !z_pred)
						  axis1
						  axis2);
				x_pred := x;
				y_pred := y;
				z_pred := z;
 				incr indice
			   ) objList;
		 
		 (* Calcule 2nde partie de la translation pour passer d'un objet au suivant *)
		 let sumCol   = (sumTab col)   +. (float_of_int (ratio1-1)) *. space1
		 and sumRow   = (sumTab row)   +. (float_of_int (ratio2-1)) *. space2
		 and sumStage = (sumTab stage) +. (float_of_int (ratio3-1)) *. space3
		 in
		   (match objList
		    with
		      | []            -> ()
		      | (_, vect)::tl -> vect#updateAll (permut ((col.(0)   -.   sumCol) /. 2.0 +. (sbt2 align1 0.0   col.(0)),
								 (row.(0)   -.   sumRow) /. 2.0 +. (sbt2 align2 0.0   row.(0)),
								 (stage.(0) -. sumStage) /. 2.0 +. (sbt2 align3 0.0 stage.(0)))
							   axis1
							   axis2);
			  indice := 1;
			  List.iter (function (_, vect) -> let i     =  !indice                 mod ratio1 
							   and i_old = (!indice -1)             mod ratio1 
							   and j     =  (!indice      / ratio1) mod ratio2
							   and j_old = ((!indice - 1) / ratio1) mod ratio2
							   and k     =  !indice     / (ratio1 * ratio2)
							   and k_old = (!indice -1) / (ratio1 * ratio2)
				     in
				     let delta = (match (i, j) 
						  with
						    | (0, 0) -> ((col.(ratio1-1) +. col.(0)) /. 2.0 -. sumCol,
								 (row.(ratio2-1) +. row.(0)) /. 2.0 -. sumRow,
								 (stage.(k-1) +. stage.(k)) /. 2.0 +. space3)
							
						    | (0, _) -> ((col.(ratio1-1) +. col.(0)) /. 2.0 -. sumCol,
								 (row.(j-1) +. row.(j)) /. 2.0 +. space2,
								 0.0)
							
						    | (_, _) -> ((col.(i-1) +. col.(i)) /. 2.0 +. space1,
								 0.0,
								 0.0)
						 )
				     and delta2 = (sbt2 align1   col.(i_old) col.(i),
						   sbt2 align2   row.(j_old) row.(j),
						   sbt2 align3 stage.(k_old) stage.(k))
				     in
				       vect#updateAll (permut (add_triplet delta delta2)
							 axis1
							 axis2);
				       incr indice
				    ) tl
		   );
		   boundingBox_of pt sumCol sumRow sumStage axis1 axis2
		     
	   | Hexa2 (axis1, axis2, align3, space, ratio1, ratio2) ->
	       let cote   = ref 0.0 
	       and indice = ref 0
	       in
	       let (x_pred, y_pred, z_pred) = (ref 0.0, ref 0.0, ref 0.0)
	       in
		 (* Calcule de la taille de chaque case de l'HexaGri2 et de la 1re partie de la translation pour passer d'un objet au suivant *)
		 List.iter (function (obj, vect) -> let (min1, min2, min3, max1, max2, max3) = desordre (boundingBox obj) axis1 axis2
						    and i = !indice mod ratio1 
						    and j = !indice / ratio1 
			    in
			      max_diff_update_ref cote max1 min1;
			      max_diff_update_ref cote max2 min2;
			      max_diff_update_ref cote max3 min3;
			      let (x, y, z) = (sbt Middle min1 max1,
					       sbt Middle min2 max2,
					       sbt align3 min3 max3)
			      in
				vect#setAll (permut (x -. !x_pred,
						     y -. !y_pred,
						     z -. !z_pred)
					       axis1
					       axis2);
				x_pred := x;
				y_pred := y;
				z_pred := z;
 				incr indice
			   ) objList;
		 
		 (* Calcule 2nde partie de la translation pour passer d'un objet au suivant *)
  		 let rayon  = (!cote /. sqrt3) +. (space /. 2.0)
		 in
		 let sumCol = (2.0 *. rayon *. (float_of_int (ratio1-1))) +. !cote +. (if ((ratio2 > 1)
											   && ((length objList) >= (2 * ratio1))
											  )
										       then rayon
										       else 0.0)
		 and sumRow = ((float_of_int (ratio2-1)) *. rayon *. sqrt3) +. !cote
		 in
		   (match objList
		    with
		      | []            -> ()
	 	      | (_, vect)::tl -> vect#updateAll (permut ((!cote -. sumCol) /. 2.0, 
								 (!cote -. sumRow) /. 2.0,
								 sbt2 align3 0.0 !cote)
							   axis1
							   axis2);
			  indice := 1;
			  List.iter (function (_, vect) -> let i   = (!indice)     mod ratio1
							   and j   = (!indice)     / ratio1
				     in
				     let sup = if (j mod 2)=0
				     then (-. rayon)
				     else (   rayon)
				     in
				     let (delta1, delta2) = (if (i <> 0) 
							     then (2.0 *. rayon,
								   0.0)
							     else ((2.0 *. rayon *. (float_of_int (1-ratio1))) +. sup,
								   rayon *. sqrt3)
							    )
				     in
				       vect#updateAll (permut (delta1, delta2, 0.0)
							 axis1
							 axis2);
 				       incr indice
				    ) tl
		   );
		   boundingBox_of pt sumCol sumRow !cote axis1 axis2
		     
	   | Hexa3 (axis1, axis2, space, ratio1, ratio2, ratio3) ->
	       let cote   = ref 0.0
	       and indice = ref 0 
	       in 
	       let (x_pred, y_pred, z_pred) = (ref 0.0, ref 0.0, ref 0.0)
	       in
		 (* Calcule de la taille de chaque case de l'HexaGrid3 et de la 1re partie de la translation pour passer d'un objet au suivant *)
		 List.iter (function (obj, vect) -> let (min1, min2, min3, max1, max2, max3) = desordre (boundingBox obj) axis1 axis2
						    and i = !indice mod ratio1
						    and j = (!indice / ratio1) mod ratio2
						    and k = !indice / (ratio1 * ratio2) 
			    in
			      max_diff_update_ref cote max1 min1;
			      max_diff_update_ref cote max2 min2;
			      max_diff_update_ref cote max3 min3;
			      let (x, y, z) = (sbt Middle min1 max1,
					       sbt Middle min2 max2,
					       sbt Middle min3 max3)
			      in
				vect#updateAll (permut (x -. !x_pred,
							y -. !y_pred,
							z -. !z_pred)
						  axis1
						  axis2);
				x_pred := x;
				y_pred := y;
				z_pred := z;
 				incr indice
			   ) objList;
		 
		 (* Calcule 2nde partie de la translation pour passer d'un objet au suivant *)
		 let n      = length objList
  		 and rayon  = (!cote *. sqrt6 /. 4.0) +. (space /. 2.0)
		 in
		 let sumCol = (2.0 *. rayon *. (float_of_int (ratio1-1))) +. !cote +. (if ((ratio2 > 1)
											   && (n >= (2 * ratio1))
											  )
										       then rayon
										       else 0.0)
		 and sumRow = ((float_of_int (ratio2-1)) *. rayon *. sqrt3) +. !cote +. (if ((ratio2 > 1)
											     && (n > ratio1 * (2 * ratio1 - 1))
											    )
											 then 2.0 *. rayon /. sqrt3
											 else 0.0)
		 and sumStage = (((float_of_int (ratio3-1)) *. 2.0 *. rayon *. sqrt6) /. 3.0) +. !cote
		 in
		   (match objList
		    with
		      | []            -> ()
		      | (_, vect)::tl -> vect#updateAll (permut ((!cote -.   sumCol) /. 2.0,
								 (!cote -.   sumRow) /. 2.0,
								 (!cote -. sumStage) /. 2.0)
							   axis1
							   axis2);
			  indice := 1;
			  List.iter (function (_, vect) -> let i     =  !indice                 mod ratio1 
							   and j     =  (!indice      / ratio1) mod ratio2
							   and k     =  !indice     / (ratio1 * ratio2)
				     in
				     let sup11 = if (ratio2 mod 2)=0
				     then (-. rayon)
				     else 0.0
				     and sup12 = if (j mod 2)=0
				     then (-. rayon)
				     else (   rayon)
				     and sup2  = if (k mod 2)=0
				     then (-. rayon *. 2.0 /. sqrt3)
				     else (   rayon *. 2.0 /. sqrt3)
				     in
				     let delta = (match (i, j) 
						  with
						    | (0, 0) -> ((2.0 *. rayon *. (float_of_int (1-ratio1))) +. sup11,
								 (sqrt3 *. rayon *. (float_of_int (1-ratio2))) +. sup2,
								 2.0 *. sqrt6 *. rayon /. 3.0)
							
						    | (0, _) -> ((2.0 *. rayon *. (float_of_int (1-ratio1))) +. sup12,
								 rayon *. sqrt3,
								 0.0)
							
						    | (_, _) -> (2.0 *. rayon,
								 0.0,
								 0.0)
						 )
				     in
				       vect#updateAll (permut delta axis1 axis2);
				       incr indice
				    ) tl
		   );
		   boundingBox_of pt sumCol sumRow sumStage axis1 axis2
	)
	
	
	
(********************************************************)
(* Calcule le BoundingBox d'une scne scn               *)
(********************************************************)
(* Type : scene ->                                      *)
(*        float * float * float * float * float * float *)
(********************************************************)
let boundingScene scn = match scn#getContenu
			with
			  | []     -> (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
			  | hd::tl -> _boundingList tl (boundingBox hd)
