(****************************************************)
(*                                                  *)
(* Ce fichier contient les fonctions ncessaires au *)
(* dessin de chacunes des primitives disponibles.   *)
(*                                                  *)
(****************************************************)



open Types
open Parametres



(************************************************)
(* Fixe la normale  norm ou  son oppos selon *)
(* drawInt.                                     *)
(************************************************)
(* Type : float * float * float -> bool -> unit *)
(************************************************)
let normeSens norm drawInt = GlDraw.normal3 (if drawInt
					     then (Triplet.oppose_triplet norm)
					     else norm)



(**********************************************************)
(* Dessine une bote (ou un morceau de bote) en fonction *)
(* de la 1/2 taille sur chaque axe.                       *)
(* secteur  dtermine  quel  morceau  de  bote  il  faut *)
(* dessiner :                                             *)
(*   - Inf    -> 1/2 bote infrieur                      *)
(*   - Sup    -> 1/2 bote suprieur                      *)
(*   - All    -> bote entire                            *)
(*   - Bottom -> Fond de la bote                         *)
(* drawInt indique si l'intrieur est galement dessin.  *)
(**********************************************************)
(* Type : float -> float -> float -> Types.cont_mode ->   *)
(*        bool -> unit                                    *)
(**********************************************************)
let box x y z secteur drawInt = let drawSol drawInt = normeSens (0., -1., 0.) drawInt;
						      GlDraw.vertex3 (  x, -.y,  z);
						      GlDraw.vertex3 (-.x, -.y,  z);
						      GlDraw.vertex3 (-.x, -.y,-.z);
						      GlDraw.vertex3 (  x, -.y,-.z)

				and drawToit drawInt = normeSens (0., 1., 0.) drawInt;
						       GlDraw.vertex3 (  x, y,  z);
						       GlDraw.vertex3 (  x, y,-.z);
						       GlDraw.vertex3 (-.x, y,-.z);
						       GlDraw.vertex3 (-.x, y,  z)

				and drawMurs y1 y2 drawInt = let un = if drawInt
								      then -1.
								      else  1.
							     in
							       GlDraw.normal3 (  un, 0.,  0.);
							       GlDraw.vertex3 (   x, y2,   z);
							       GlDraw.vertex3 (   x, y1,   z);
							       GlDraw.vertex3 (   x, y1,-. z);
							       GlDraw.vertex3 (   x, y2,-. z);

							       GlDraw.normal3 (-.un, 0.,  0.);
							       GlDraw.vertex3 (-. x, y2,   z);
							       GlDraw.vertex3 (-. x, y2,-. z);
							       GlDraw.vertex3 (-. x, y1,-. z);
							       GlDraw.vertex3 (-. x, y1,   z);

							       GlDraw.normal3 (  0., 0.,  un);
							       GlDraw.vertex3 (   x, y2,   z);
							       GlDraw.vertex3 (-. x, y2,   z);
							       GlDraw.vertex3 (-. x, y1,   z);
							       GlDraw.vertex3 (   x, y1,   z);

							       GlDraw.normal3 (  0., 0.,-.un);
							       GlDraw.vertex3 (   x, y2,-. z);
							       GlDraw.vertex3 (   x, y1,-. z);
							       GlDraw.vertex3 (-. x, y1,-. z);
							       GlDraw.vertex3 (-. x, y2,-. z)
				in
				let sbt = match secteur
					  with
					    | Bottom -> (fun drawInt -> drawSol drawInt)

					    | Inf    -> (fun drawInt -> drawSol drawInt;
									drawMurs (-.y) 0. drawInt)

					    | Sup    -> (fun drawInt -> drawToit drawInt;
									drawMurs 0. y drawInt)

					    | All    -> (fun drawInt -> drawSol drawInt;
									drawToit drawInt;
									drawMurs (-.y) y drawInt)
				in
				  GlDraw.begins `quads;
				  sbt false;
				  GlDraw.ends ();
				  if drawInt
				  then
				    begin
				      GlDraw.front_face `cw;
				      GlDraw.begins `quads;
				      sbt true;
				      GlDraw.ends ();
				      GlDraw.front_face `ccw
				    end



(***************************************************************)
(* Dessine une sphre (ou un morceau de sphre) en fonction du *)
(* nombre de  mridiens (slice),  de parallles (stack)  et du *)
(* rayon (radius).                                             *)
(* secteur dtermine quel morceau de sphre il faut dessiner : *)
(*   - Inf    -> 1/2 sphre infrieur                          *)
(*   - Sup    -> 1/2 sphre suprieur                          *)
(*   - All    -> sphre entire                                *)
(*   - Bottom -> Ple Sud                                      *)
(* drawInt indique si l'intrieur est galement dessin.       *)
(***************************************************************)
(* Type : int -> int -> float -> Types.cont_mode -> bool ->    *)
(*        unit                                                 *)
(***************************************************************)
let sphere slice stack radius secteur drawInt = let (min, max) = match secteur
								 with
								   | Inf    -> (1, (stack+1)/2)
								   | Sup    -> (stack/2, stack-1)
								   | All    -> (1, stack-1)
								   | Bottom -> (1, (stack+4)/5)
						in
						let delta = max - min
						in
						let cosStack = Array.make (delta + 1) 0.
						and sinStack = Array.make (delta + 1) 0.
						and cosSlice = Array.make slice 0.
						and sinSlice = Array.make slice 0.
						in
						  for i =  0  to delta
						  do
						    let theta = ((pi *. (float_of_int (i+min))) /. (float_of_int stack)) -. Parametres.pi_2
						    in
						      cosStack.(i) <- cos theta;
						      sinStack.(i) <- sin theta
						  done;
						  (match secteur
						   with 
						     | Inf    -> cosStack.(delta) <- 1.;
								 sinStack.(delta) <- 0.
						     | Sup    -> cosStack.(0) <- 1.;
								 sinStack.(0) <- 0.
						     | Bottom -> cosStack.(delta) <- 0.5877852522924731;
								 sinStack.(delta) <- -0.8090169943749474
						     | _      -> ()
						  );

						  for j =  0  to (slice - 1)
						  do
						    let phi = (pi2 *. (float_of_int j)) /. (float_of_int slice)
						    in
						      cosSlice.(j) <- cos phi;
						      sinSlice.(j) <- sin phi
						  done;

						  let drawSol drawInt = GlDraw.begins `triangle_fan;
									normeSens (0., -1., 0.) drawInt;
									GlDraw.vertex3 (0., -.radius, 0.);
									for j = 0 to (slice - 1)
									do
									  normeSens (cosSlice.(j) *. cosStack.(0), sinStack.(0), sinSlice.(j) *. cosStack.(0))
										    drawInt;
									  GlDraw.vertex3 (radius *. cosSlice.(j) *. cosStack.(0), radius *. sinStack.(0), radius *. sinSlice.(j) *. cosStack.(0))
									done;
									normeSens (cosStack.(0), sinStack.(0), 0.) drawInt;
									GlDraw.vertex3 (radius *. cosStack.(0), radius *. sinStack.(0), 0.);
									GlDraw.ends ()

						  and drawToit drawInt = GlDraw.begins `triangle_fan;
									 normeSens (0., 1., 0.) drawInt;
									 GlDraw.vertex3 (0., radius, 0.);
									 normeSens (cosStack.(delta), sinStack.(delta), 0.) drawInt;
									 GlDraw.vertex3 (radius *. cosStack.(delta), radius *. sinStack.(delta), 0.);
									 for j = (slice - 1) downto 0
									 do
									   normeSens (cosSlice.(j) *. cosStack.(delta), sinStack.(delta), sinSlice.(j) *. cosStack.(delta))
										     drawInt;
									   GlDraw.vertex3 (radius *. cosSlice.(j) *. cosStack.(delta), radius *. sinStack.(delta), radius *. cosStack.(delta) *. sinSlice.(j))
									 done;
									 GlDraw.ends ()

						  and drawCote drawInt = GlDraw.begins `quad_strip;
									 for i = 0 to (delta - 1)
									 do
									   for j = 0 to (slice - 1)
									   do
									     normeSens (cosSlice.(j) *. cosStack.(i), sinStack.(i), sinSlice.(j) *. cosStack.(i))
										       drawInt;
									     GlDraw.vertex3 (radius *. cosSlice.(j) *. cosStack.(i), radius *. sinStack.(i), radius *. cosStack.(i) *. sinSlice.(j));
									     normeSens (cosSlice.(j) *. cosStack.(i+1), sinStack.(i+1), sinSlice.(j) *. cosStack.(i+1))
										       drawInt;
									     GlDraw.vertex3 (radius *. cosSlice.(j) *. cosStack.(i+1), radius *. sinStack.(i+1), radius *. cosStack.(i+1) *. sinSlice.(j))
									   done;
									   normeSens (cosStack.(i), sinStack.(i), 0.0) drawInt;
									   GlDraw.vertex3 (radius *. cosStack.(i), radius *. sinStack.(i), 0.0);
									   normeSens (cosStack.(i+1), sinStack.(i+1), 0.0) drawInt;
									   GlDraw.vertex3 (radius *. cosStack.(i+1), radius *. sinStack.(i+1), 0.0)
									 done;
									 GlDraw.ends ()
						  in

						  let sbt = match secteur
							    with
							      | Bottom -> (fun drawInt -> drawSol drawInt)

							      | Inf    -> (fun drawInt -> drawSol drawInt;
											  drawCote drawInt)

							      | Sup    -> (fun drawInt -> drawToit drawInt;
											  drawCote drawInt)

							      | All    -> (fun drawInt -> drawSol drawInt;
											  drawToit drawInt;
											  drawCote drawInt)
						  in
						    sbt false;
						    if drawInt
						    then
						      begin
							GlDraw.front_face `cw;
							sbt true;
							GlDraw.front_face `ccw
						      end



(***********************************************************************)
(* Dessine une ellipsode (ou un morceau d'ellipsode) en fonction du  *)
(* nombre de  mridiens (slice),  de parallles (stack)  et des rayons *)
(* sur chaque axe.                                                     *)
(* secteur dtermine quel morceau de sphre il faut dessiner :         *)
(*   - Inf    -> 1/2 sphre infrieur                                  *)
(*   - Sup    -> 1/2 sphre suprieur                                  *)
(*   - All    -> sphre entire                                        *)
(*   - Bottom -> Ple Sud                                              *)
(***********************************************************************)
(* Type : int -> int -> float -> float -> float -> cont_mode -> unit   *)
(***********************************************************************)
let ellipsoide slice stack rx ry rz secteur drawInt = GlMat.push ();
						      GlMat.scale3 (rx, ry, rz);
						      sphere slice stack 1.0 secteur drawInt;
						      GlMat.pop ()



(*******************************************)
(* Dessine un disque en fonction du nombre *)
(* de cts (slice) et du rayon (radius).  *)
(*******************************************)
(* Type : int -> float -> unit             *)
(*******************************************)
let disc slice radius = let cosSlice = Array.make slice 0.0
			and sinSlice = Array.make slice 0.0
			in
			  for j =  0  to (slice - 1)
			  do
			    let phi = (pi2 *. (float_of_int j)) /. (float_of_int slice)
			    in
			      cosSlice.(j) <- cos phi;
			      sinSlice.(j) <- sin phi
			  done;

			  (* Ple Nord *)
			  GlDraw.begins `triangle_fan;
			  GlDraw.normal3 (0.0, 1.0, 0.0);
			  GlDraw.vertex3 (0.0, 0.0, 0.0);
			  GlDraw.vertex3 (radius, 0.0, 0.0);
			  for j = (slice - 1) downto 0
			  do
			    GlDraw.vertex3 (radius *. cosSlice.(j), 0.0, radius *. sinSlice.(j))
			  done;
			  GlDraw.ends ();

			  (* Ple Sud *)
			  GlDraw.begins `triangle_fan;
			  GlDraw.normal3 (0.0, -1.0, 0.0);
			  GlDraw.vertex3 (0.0, 0.0, 0.0);
			  for j = 0 to (slice - 1)
			  do
			    GlDraw.vertex3 (radius *. cosSlice.(j), 0.0, radius *. sinSlice.(j))
			  done;
			  GlDraw.vertex3 (radius, 0.0, 0.0);
			  GlDraw.ends ()



(***************************************************************)
(* Dessine un cylindre en fonction du nombre de cts (slice), *)
(* du rayon (radius) et de la demi-hauteur (height).           *)
(* solid indique si le cylindre est plein.                     *)
(* secteur dtermine quel morceau de sphre il faut dessiner : *)
(*   - Inf    -> 1/2 sphre infrieur                          *)
(*   - Sup    -> 1/2 sphre suprieur                          *)
(*   - All    -> sphre entire                                *)
(*   - Bottom -> Ple Sud                                      *)
(* drawInt indique si l'intrieur est galement dessin.       *)
(***************************************************************)
(* Type : int -> float -> float -> bool -> Types.cont_mode ->  *)
(*        bool -> unit                                         *)
(***************************************************************)
let cylinder slice radius height solid secteur drawInt = let cosSlice   = Array.make slice 0.0
							 and sinSlice   = Array.make slice 0.0
							 in
							   for j =  0  to (slice - 1)
							   do
							     let phi = (pi2 *. (float_of_int j)) /. (float_of_int slice)
							     in
							       cosSlice.(j) <- cos phi;
							       sinSlice.(j) <- sin phi
							   done;

							 let haut drawInt = GlDraw.begins `triangle_fan;
									    normeSens (0.0, 1.0, 0.0) drawInt;
									    GlDraw.vertex3 (0.0, height, 0.0);
									    GlDraw.vertex3 (radius, height, 0.0);
									    for j = (slice - 1) downto 0
									    do
									      GlDraw.vertex3 (radius *. cosSlice.(j), height, radius *. sinSlice.(j))
									    done;
									    GlDraw.ends ()

							 and bas drawInt = GlDraw.begins `triangle_fan;
									   normeSens (0.0, -1.0, 0.0) drawInt;
									   GlDraw.vertex3 (0.0, -.height, 0.0);
									   for j = 0 to (slice - 1)
									   do
									     GlDraw.vertex3 (radius *. cosSlice.(j), -.height, radius *. sinSlice.(j))
									   done;
									   GlDraw.vertex3 (radius, -.height, 0.0);
									   GlDraw.ends ()

							 and cote y1 y2 drawInt = GlDraw.begins `quad_strip;
										  for j = 0 to (slice - 1)
										  do
										    normeSens (cosSlice.(j), 0.0, sinSlice.(j)) drawInt;
										    GlDraw.vertex3 (radius *. cosSlice.(j), y1, radius *. sinSlice.(j));
										    GlDraw.vertex3 (radius *. cosSlice.(j), y2, radius *. sinSlice.(j))
										  done;
										  normeSens (cosSlice.(0), 0.0, sinSlice.(0)) drawInt;
										  GlDraw.vertex3 (radius, y1, 0.0);
										  GlDraw.vertex3 (radius, y2, 0.0);
										  GlDraw.ends ()
							 in

							 let sbt = match solid, secteur
								   with
								     |  true, Bottom -> (fun drawInt -> bas drawInt)

								     |  true, Inf    -> (fun drawInt -> bas drawInt;
													cote (-.height) 0. drawInt)

								     |  true, Sup    -> (fun drawInt -> haut drawInt;
													cote 0. height drawInt)

								     |  true, All    -> (fun drawInt -> bas drawInt;
													haut drawInt;
													cote (-.height) height drawInt)

								     | false, _      -> (fun drawInt -> cote (-.height) height drawInt)
							 in
							   sbt false;
							   if drawInt
							   then
							     begin
							       GlDraw.front_face `cw;
							       sbt true;
							       GlDraw.front_face `ccw
							     end



(***********************************************************)
(* Dessine un cne en fonction du nombre de cts (slice), *)
(* du rayon (radius) et de la composante sur Y de la base  *)
(* (down) et du sommet (up).                               *)
(* solid indique si le cne est plein.                     *)
(* drawInt indique si l'intrieur est galement dessin.   *)
(***********************************************************)
(* Type : int -> float -> float -> float -> bool ->        *)
(*        bool -> unit                                     *)
(***********************************************************)
let _cone slice radius down up solid drawInt = let height = up -. down
					       in
					       let cosSlice = Array.make slice 0.0
					       and sinSlice = Array.make slice 0.0
					       and cst = 2.0 *. (cos (pi /. (float_of_int slice)))
					       and yNormale = 1.0 /. (sqrt (1.0 +. 4.0*.height*.height/.(radius*.radius)))
					       in
					       let cst2 = 2.0 *. height *. yNormale /. radius
					       in
					       let triangle i1 i2 drawInt = let normale sinT cosT = normeSens (cst2 *. cosT, yNormale, cst2 *. sinT) drawInt
									    in
									      normale sinSlice.(i1) cosSlice.(i1);
									      GlDraw.vertex3 (radius *. cosSlice.(i1), down, radius *. sinSlice.(i1));
									      normale ((sinSlice.(i1) +. sinSlice.(i2)) /. cst) ((cosSlice.(i1) +. cosSlice.(i2)) /. cst);
									      GlDraw.vertex3 (0.0, up, 0.0);
									      normale sinSlice.(i2) cosSlice.(i2);
									      GlDraw.vertex3 (radius *. cosSlice.(i2), down, radius *. sinSlice.(i2))
					       in
						 for j = 0  to (slice - 1)
						 do
						   let phi = (pi2 *. (float_of_int j)) /. (float_of_int slice)
						   in
						     cosSlice.(j) <- cos phi;
						     sinSlice.(j) <- sin phi
						 done;

						 let cote drawInt = GlDraw.begins `triangles;
								    for j = 0 to (slice-2)
								    do
								      triangle j (j+1) drawInt
								    done;
								    triangle (slice-1) 0 drawInt;
								    GlDraw.ends ()

						 and bas drawInt = GlDraw.begins `triangle_fan;
								   normeSens (0.0, -1.0, 0.0) drawInt;
								   GlDraw.vertex3 (0.0, down, 0.0);
								   for j = 0 to (slice - 1)
								   do
								     GlDraw.vertex3 (radius *. cosSlice.(j), down, radius *. sinSlice.(j))
								   done;
								   GlDraw.vertex3 (radius, down, 0.0);
								   GlDraw.ends ()
						 in

						 let sbt = match solid
							   with
							     |  true -> (fun drawInt -> bas drawInt;
											cote drawInt)

							     | false -> (fun drawInt -> cote drawInt)
						 in
						   sbt false;
						   if drawInt
						   then
						     begin
						       GlDraw.front_face `cw;
						       sbt true;
						       GlDraw.front_face `ccw
						     end



(***********************************************************)
(* Dessine un tronc de cne en fonction du nombre de cts *)
(* (slice), du rayon (radius), de la composante sur Y de   *)
(* la base (down) et du sommet (up) et du rapport entre    *)
(* les rayons suprieurs et infrieur.                     *)
(* solidBas et solidHaut indiquent si la base et le sommet *)
(* du tronc de cne sont pleins.                           *)
(* drawInt indique si l'intrieur est galement dessin.   *)
(***********************************************************)
(* Type : int -> float -> float -> float -> float ->       *)
(*        bool -> bool -> unit                             *)
(***********************************************************)
let _frustum slice radius down up rate solidBas solidHaut drawInt = let height = up -. down
								    in
								    let cosSlice  = Array.make slice 0.0
								    and sinSlice  = Array.make slice 0.0
								    and rayonHaut = radius *. rate
								    and cst = 2.0 *. (cos (pi /. (float_of_int slice)))
								    and yNormale = 1.0 /. (sqrt (1.0 +. 4.0*.height*.height*.rate*.rate/.(radius*.radius)))
								    in
								    let cst2 = 2.0 *. height *. rate *. yNormale /. radius
								    in
								    let trait i = normeSens (cst2 *. cosSlice.(i), yNormale, cst2 *. sinSlice.(i)) drawInt;
										  GlDraw.vertex3 (radius    *. cosSlice.(i), down, radius    *. sinSlice.(i));
										  GlDraw.vertex3 (rayonHaut *. cosSlice.(i),   up, rayonHaut *. sinSlice.(i));
								    in
								      for j =  0  to (slice - 1)
								      do
									let phi = (pi2 *. (float_of_int j)) /. (float_of_int slice)
									in
									  cosSlice.(j) <- cos phi;
									  sinSlice.(j) <- sin phi
								      done;

								      let cote drawInt = GlDraw.begins `quad_strip;
											 for j = 0 to (slice-1)
											 do
											   trait j
											 done;
											 trait 0;
											 GlDraw.ends ()

								      and haut drawInt = GlDraw.begins `triangle_fan;
											 normeSens (0.0,  1.0, 0.0) drawInt;
											 GlDraw.vertex3 (0.0, up, 0.0);
											 GlDraw.vertex3 (rayonHaut, up, 0.0);
											 for j = (slice - 1) downto 0
											 do
											   GlDraw.vertex3 (rayonHaut *. cosSlice.(j), up, rayonHaut *. sinSlice.(j))
											 done;
											 GlDraw.ends ()

								      and bas drawInt = GlDraw.begins `triangle_fan;
											normeSens (0.0, -1.0, 0.0) drawInt;
											GlDraw.vertex3 (0.0, down, 0.0);
											for j = 0 to (slice - 1)
											do
											  GlDraw.vertex3 (radius *. cosSlice.(j), down, radius *. sinSlice.(j))
											done;
											GlDraw.vertex3 (radius, down, 0.0);
											GlDraw.ends ()
								      in

								      let sbt = match solidHaut, solidBas
										with
										  |  true, true  -> (fun drawInt -> bas drawInt;
														    haut drawInt;
														    cote drawInt)

										  |  true, false -> (fun drawInt -> haut drawInt;
														    cote drawInt)

										  | false,  true -> (fun drawInt -> bas drawInt;
														    cote drawInt)

										  | false, false -> (fun drawInt -> cote drawInt)
								      in
									sbt false;
									if drawInt
									then
									  begin
									    GlDraw.front_face `cw;
									    sbt true;
									    GlDraw.front_face `ccw
									  end



(***************************************************************)
(* Dessine la primitive Empty.                                 *)
(* secteur dtermine quel morceau de sphre il faut dessiner : *)
(*   - Inf    -> 1/2 sphre infrieur                          *)
(*   - Sup    -> 1/2 sphre suprieur                          *)
(*   - All    -> sphre entire                                *)
(*   - Bottom -> Ple Sud                                      *)
(* drawInt sert juste  rendre le typage compatible avec les   *)
(* fonctions d'affichage des autres primitives volumiques.     *)
(***************************************************************)
(* Type : float -> float -> float -> Types.cont_mode ->        *)
(*        bool -> unit                                         *)
(***************************************************************)
let empty x y z secteur drawInt = (* Les traits n'ont pas de normale -> on dsactive la lumire *)
				  Gl.disable `lighting;
				  if !Parametres.polygon_mode = `point
				  then
				    begin
				      GlDraw.begins `points;
				      let bas  () = GlDraw.vertex3 (-. x, -.y,   z);
						    GlDraw.vertex3 (-. x, -.y, -.z);
						    GlDraw.vertex3 (   x, -.y,   z);
						    GlDraw.vertex3 (   x, -.y, -.z)

				      and haut () = GlDraw.vertex3 (  x, y, -.z);
						    GlDraw.vertex3 (  x, y,   z);
						    GlDraw.vertex3 (-.x, y, -.z);
						    GlDraw.vertex3 (-.x, y,   z)
				      in
					(match secteur
					 with
					   | Bottom
					   | Inf -> bas ()
					   | Sup -> haut ()
					   | All -> bas ();
						    haut ()
					);
					GlDraw.ends ()
				    end
				  else
				    begin
				      GlDraw.line_stipple ~factor:5 0xAAAA;
				      let bas  () = GlDraw.begins `line_strip;
						    GlDraw.vertex3 (  x, -.y,   z);
						    GlDraw.vertex3 (  x, -.y, -.z);
						    GlDraw.vertex3 (-.x, -.y, -.z);
						    GlDraw.vertex3 (-.x, -.y,   z);
						    GlDraw.vertex3 (  x, -.y,   z);
						    GlDraw.ends ()

				      and haut () = GlDraw.begins `line_strip;
						    GlDraw.vertex3 (  x, y,   z);
						    GlDraw.vertex3 (  x, y, -.z);
						    GlDraw.vertex3 (-.x, y, -.z);
						    GlDraw.vertex3 (-.x, y,   z);
						    GlDraw.vertex3 (  x, y,   z);
						    GlDraw.ends ()

				      and cote y1 y2 = GlDraw.begins `lines;
						       GlDraw.vertex3 (  x, y1,   z);
						       GlDraw.vertex3 (  x, y2,   z);
						       GlDraw.vertex3 (-.x, y1,   z);
						       GlDraw.vertex3 (-.x, y2,   z);
						       GlDraw.vertex3 (  x, y1, -.z);
						       GlDraw.vertex3 (  x, y2, -.z);
						       GlDraw.vertex3 (-.x, y1, -.z);
						       GlDraw.vertex3 (-.x, y2, -.z);
						       GlDraw.ends ()
				      in
				      let sbt () = match secteur
						   with
						     | Bottom -> bas ()
						     | Inf    -> bas ();
								 cote (-.y) 0.
						     | Sup    -> haut ();
								 cote 0. y
						     | All    -> bas  ();
								 haut ();
								 cote (-.y) y
				      in
					if !Parametres.isVectoring 
					then
					  begin
					    (* Pointills :) *)
					    Vectoriel.psEnable `line_stipple;
					    sbt ();
					    Vectoriel.psDisable `line_stipple
					  end
					else
					  begin
					    Gl.enable `line_stipple;
					    sbt ();
					    Gl.disable `line_stipple
					  end;
				    end;
				  if !Parametres.lighting
				  then
				    Gl.enable `lighting



(**************************************************)
(* Dessine un polygone  partir d'une liste       *)
(* ORDONNEE de points et de sa normale.           *)
(**************************************************)
(* Type : Types.point list -> Types.point -> unit *)
(**************************************************)
let polygone points normale = 
  if (!Parametres.withZbug)
  then 
    begin
      (* Front *)
      GlDraw.begins `polygon;
      GlDraw.normal3 (let (x,y,z) = (Triplet.oppose_triplet normale#getAll) in (x,y,-.z));
      List.iter (function pt -> GlDraw.vertex3 (pt#getX, pt#getY, -.pt#getZ)) points;
      GlDraw.ends ();
      (* Back *)
      GlDraw.begins `polygon;
      GlDraw.normal3 (let (x,y,z) = normale#getAll in (x,y,-.z));
      Usefull.revIter (function pt -> GlDraw.vertex3 (pt#getX, pt#getY, -.pt#getZ)) points;
      GlDraw.ends ()
    end
  else
    begin
      (* Front *)
      GlDraw.begins `polygon;
      GlDraw.normal3 normale#getAll;
      List.iter (function pt -> GlDraw.vertex3 (pt#getX, pt#getY, pt#getZ)) points;
      GlDraw.ends ();
      (* Back *)
      GlDraw.begins `polygon;
      GlDraw.normal3 (Triplet.oppose_triplet normale#getAll);
      Usefull.revIter (function pt -> GlDraw.vertex3 (pt#getX, pt#getY, pt#getZ)) points;
      GlDraw.ends ()
    end


(***********************************)
(* Dessine un polygone  partir    *)
(* d'une liste de points.          *)
(***********************************)
(* Type : Types.point list -> unit *)
(***********************************)
let polyline points = let sbt shape = GlDraw.begins shape;
				      List.iter (function pt -> GlDraw.vertex3 (pt#getX, pt#getY, if (!Parametres.withZbug) then -.pt#getZ else pt#getZ)) points;
				      GlDraw.ends ()
		      in
			(* Les traits n'ont pas de normale -> on dsactive la lumire *)
			Gl.disable `lighting;
			if !Parametres.polygon_mode = `point
			then
			    sbt `points
			else
			  begin
			    if !Parametres.isVectoring
			    then
			      begin
				Vectoriel.psLineWidth 1.5;
				sbt `line_strip;
				Vectoriel.psLineWidth 1.0
			      end
			    else
			      begin
				GlDraw.line_width 1.5;
				sbt `line_strip;
				GlDraw.line_width 1.0
			      end
			  end;
			if !Parametres.lighting
			then
			  Gl.enable `lighting



(***************************************************************)
(* Dessine un cne en fonction du nombre de cts (slice), du  *)
(* rayon (radius) et de la demi-hauteur (height).              *)
(* solid indique si le cne est plein.                         *)
(* secteur dtermine quel morceau de sphre il faut dessiner : *)
(*   - Inf    -> 1/2 sphre infrieur                          *)
(*   - Sup    -> 1/2 sphre suprieur                          *)
(*   - All    -> sphre entire                                *)
(*   - Bottom -> Ple Sud                                      *)
(* drawInt indique si l'intrieur est galement dessin.       *)
(***************************************************************)
(* Type : int -> float -> float -> bool -> Types.cont_mode ->  *)
(*        unit                                                 *)
(***************************************************************)
let cone slice radius height solid secteur drawInt = match secteur
						     with
						       | Bottom -> GlMat.push ();
								   GlMat.translate3 (0., -.height, 0.);
								   disc slice radius;
								   GlMat.pop ()
						       | Inf    -> _frustum slice radius (-.height) 0. 0.5 solid false drawInt
						       | Sup    -> _cone slice radius 0. height false drawInt
						       | All    -> _cone slice radius (-.height) height solid drawInt



(***************************************************************)
(* Dessine un tronc de cne en fonction du nombre de cts     *)
(* (slice), du rayon (radius), de la demi-hauteur (height) et  *)
(* du rapport entre les rayons suprieurs et infrieur.        *)
(* solid indique si le tronc de cne est plein.                *)
(* secteur dtermine quel morceau de sphre il faut dessiner : *)
(*   - Inf    -> 1/2 sphre infrieur                          *)
(*   - Sup    -> 1/2 sphre suprieur                          *)
(*   - All    -> sphre entire                                *)
(*   - Bottom -> Ple Sud                                      *)
(* drawInt indique si l'intrieur est galement dessin.       *)
(***************************************************************)
(* Type : int -> float -> float -> float -> Types.cont_mode -> *)
(*        bool -> unit                                         *)
(***************************************************************)
let frustum slice radius height rate solid secteur drawInt = match secteur
							     with
							       | Bottom -> GlMat.push ();
									   GlMat.translate3 (0., -.height, 0.);
									   disc slice radius;
									   GlMat.pop ()
							       | Inf    -> _frustum slice radius (-.height)     0. (0.5 *. (rate+.1.))  solid false drawInt
							       | Sup    -> _frustum slice radius         0. height (2./.(1.+.1./.rate)) false solid drawInt
							       | All    -> _frustum slice radius (-.height) height rate                 solid solid drawInt


(***************************************************************)
(* Dessine le texte (text) de taille (size) au centre de la    *)
(* scne.                                                      *)
(* secteur et drawInt sont sans intret                        *)
(***************************************************************)
(* Type : float -> string -> Types.cont_mode -> bool -> unit   *)
(***************************************************************)
let dstring pxmp secteur drawInt =
  Gl.enable `blend;
  GlPix.raster_pos ~x:0.0 ~y:0.0 ~z:0.0 ();
  GlPix.draw pxmp;
  Gl.disable `blend;
  ()
;;


(*******************************************************)
(* Dessine une primitive d'aprs ses caractristiques. *)
(*******************************************************)
(* Type : primitive -> Types.cont_mode -> bool -> unit *)
(*******************************************************)
let draw_primitive prim secteur drawInt =
  let drawInt =
    if !Parametres.polygon_mode = `fill
    then drawInt
    else false
  in
    match prim with
      | Box pt                             -> box pt#getX pt#getY pt#getZ secteur drawInt
      | Cone (r, h, solid, slice)          -> cone !slice !r !h !solid secteur drawInt
      | Cylinder (r, h, solid, slice)      -> cylinder !slice !r !h !solid secteur drawInt
      | Disc (r, slice)                    -> disc !slice !r
      | Empty (pt, visible)                ->
	  if !visible
	  then empty pt#getX pt#getY pt#getZ secteur drawInt
      | Frustum (r, h, rate, solid, slice) -> frustum !slice !r !h !rate !solid secteur drawInt
      | Polygone (l, n)                    -> polygone !l n
      | Polyline l                         -> polyline !l
      | Sphere (r, slice, stack)           -> sphere !slice !stack !r secteur drawInt
      | String pxmp                        -> dstring !pxmp secteur drawInt

