(************************************************)
(*                                              *)
(* Lexer pour l'analyse synthaxique des entres *)
(*                                              *)
(************************************************)

{

open Parser
open Types

(*---------------------------------------------------------------------*)
(* Taken from  /usr/lib/ocaml/source/parsing/lexer.mll                 *)
(* from X. Leroy & al                                                  *)
(*---------------------------------------------------------------------*)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(*---------------------------------------------------------------------*)


(* To buffer string literals *)

let initial_string_buffer = String.create 256
let string_buff = ref initial_string_buffer
let string_index = ref 0

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0

let store_string_char c =
  if !string_index >= String.length (!string_buff) then begin
    let new_buff = String.create (String.length (!string_buff) * 2) in
      String.blit (!string_buff) 0 new_buff 0 (String.length (!string_buff));
      string_buff := new_buff
  end;
  String.unsafe_set (!string_buff) (!string_index) c;
  incr string_index

let get_stored_string () =
  let s = String.sub (!string_buff) 0 (!string_index) in
  string_buff := initial_string_buffer;
  s

(* To translate escape sequences *)

let char_for_backslash =
  match Sys.os_type with
  | "Unix" | "Win32" | "Cygwin" ->
      begin function
      | 'n' -> '\010'
      | 'r' -> '\013'
      | 'b' -> '\008'
      | 't' -> '\009'
      | c   -> c
      end
  | "MacOS" ->
      begin function
      | 'n' -> '\013'
      | 'r' -> '\010'
      | 'b' -> '\008'
      | 't' -> '\009'
      | c   -> c
      end
  | x -> (raise (Erreur "unknown system type"))

let char_for_decimal_code lexbuf i =
  let c = 100 * (Char.code(Lexing.lexeme_char lexbuf i) - 48) +
           10 * (Char.code(Lexing.lexeme_char lexbuf (i+1)) - 48) +
                (Char.code(Lexing.lexeme_char lexbuf (i+2)) - 48) in  
  if ((c < 0) || (c > 255 ))
  then raise (Erreur ("Illegal escape char"))
  else Char.chr c

(* To store the position of the beginning of a string and comment *)
let string_start_pos = ref 0

(*---------------------------------------------------------------------*)
(* END of Leroy's code (to be continued in the lexer itself !)         *)
(*---------------------------------------------------------------------*)

} 



(******************************)
(* Noms d'objets ou de scnes *)
(******************************)
let name   = ['a'-'z''A'-'Z''_'](['0'-'9''a'-'z''A'-'Z''_''-''.'])*



(***********)
(* Nombres *)
(***********)
let number = ['-''+']?['0'-'9']+('.'['0'-'9']*)?



(********************)
(* Lexer principale *)
(********************)
rule token = parse

    [' ' '\t' '\r']+         { token lexbuf }
  | '\n'                     { incr Parametres.line ; token lexbuf }
  | ','	                     { TokVirgule }
  | '{'	                     { TokDebAccol }
  | '}'	                     { TokFinAccol }
  | '['	                     { TokDebCrochet }
  | ']'	                     { TokFinCrochet }
  | '<'                      { TokInf }
  | '>'	                     { TokSup }
  | ':'                      { Tok2points }

  | '"'             { reset_string_buffer();
		      let string_start = Lexing.lexeme_start lexbuf 
		      in
			string_start_pos := string_start;
			strparse lexbuf;
			lexbuf.Lexing.lex_start_pos <-	string_start - lexbuf.Lexing.lex_abs_pos;
			TokSTR (get_stored_string())
		    }

  | "Box"                    { TokBox }
  | "Cone"                   { TokCone }
  | "Cylinder"               { TokCylinder }
  | "Disc"                   { TokDisc }
  | "Frustum"                { TokFrustum }
  | "Sphere"                 { TokSphere }
  | "Polygone"               { TokPolygone }
  | "Polyline"               { TokPolyline }
  | "Empty"                  { TokEmpty }
  | "Line"                   { TokLine }
  | "String"                 { TokString }

  | "Translated"             { TokTranslated }
  | "Scaled"                 { TokScaled }
  | "AxisRotated"            { TokAxisRotated }

  | "Group"                  { TokGroup }
  | "Grid0"                  { TokGrid0 }
  | "Grid1"                  { TokGrid1 }
  | "Grid2"                  { TokGrid2 }
  | "Grid3"                  { TokGrid3 }
  | "HexaGrid2"              { TokHexa2 }
  | "HexaGrid3"              { TokHexa3 }

  | "Size"                   { TokSize }
  | "Radius"                 { TokRadius }
  | "Height"                 { TokHeight }
  | "Solid"                  { TokSolid }
  | "Slices"                 { TokSlices }
  | "Taper"                  { TokTaper }
  | "Stacks"                 { TokStacks }
  | "PointList"              { TokPointList }
  | "Text"                   { TokText }

  | "Translation"            { TokTranslation }
  | "Scale"                  { TokScale }
  | "Axis"                   { TokAxis }
  | "Angle"                  { TokAngle }

  | "Geometry"               { TokGeometry }
  | "GridList"               { TokGridList }

  | "Visible"                { TokVisible }

  | "StartPoint"             { TokStartPoint }
  | "Space"                  { TokSpace }
  | "Space1"                 { TokSpace1 }
  | "Space2"                 { TokSpace2 }
  | "Space3"                 { TokSpace3 }
  | "LineUp1"                { TokLineUp1 }
  | "LineUp2"                { TokLineUp2 }
  | "LineUp3"                { TokLineUp3 }
  | "Ratio"                  { TokRatio }

  | "Contain"                { TokContain }
  | "Draw"                   { TokDraw }

  | "Open"                   { TokOpen }
  | "Half"                   { TokHalf }
  | "Blend"                  { TokBlend }
  | "Mixed"                  { TokMixed }
  | "Wire"                   { TokWire }
  | "Close"                  { TokClose }
  | "None"                   { TokNone }
  | "Up"                     { TokUp }
  | "Down"                   { TokDown }

  | "Color"                  { TokColor }

  | "True"                   { TokBOOL(true) }
  | "true"                   { TokBOOL(true) }
  | "False"                  { TokBOOL(false) }
  | "false"                  { TokBOOL(false) }

  | "Left"                   { TokLeft }
  | "left"                   { TokLeft }

  | "Right"                  { TokRight }
  | "right"                  { TokRight }

  | "Middle"                 { TokMiddle }
  | "middle"                 { TokMiddle }

  | "Axis1"                   { TokAxis1 }
  | "Axis2"                   { TokAxis2 }

  | "AxisX"                  { TokAxisX }
  | "AxisY"                  { TokAxisY }
  | "AxisZ"                  { TokAxisZ }

  | "Definitions"            { TokDef }
  | "Scene"                  { TokScene }
  | "Show"                   { TokShow }
  | "Hide"                   { TokHide }
  | "Replace"                { TokReplace }
  | "Play"                   { TokPlay }
  | "Exec"                   { TokExec }
  | "Pause"                  { TokPause }

  | "View"                   { TokView }
  | "Auto"                   { TokAuto }
  | "Vertical"               { TokVertical }
  | "Center"                 { TokCenter }
  | "Viewer"                 { TokViewer }

  | "Movement"               { TokMove }
  | "Spin"                   { TokSpin }
  | "Rotation"               { TokRotation }
  | "DeltaX"                 { TokDeltaX }
  | "DeltaY"                 { TokDeltaY }
  | "DeltaZ"                 { TokDeltaZ }
  | "SpeedX"                 { TokSpeedX }
  | "SpeedY"                 { TokSpeedY }
  | "SpeedZ"                 { TokSpeedZ }

  (* Les commandes shell  excuter
  | '"'[^'"']*'"'            { TokCOMMAND(String.sub (Lexing.lexeme lexbuf) 1 ((Lexing.lexeme_end lexbuf) - ((Lexing.lexeme_start lexbuf) + 2))) }
*)
  (* Les noms d'objets *)
  | name                     { TokNOM(Lexing.lexeme lexbuf) }

  (* Les nombres *)
  | number                   { TokNOMBRE(float_of_string (Lexing.lexeme lexbuf)) }

  (* Les commentaires : /* ... */ ou // ... \n ou (* ... *) ou (# ... #) *)
  | "//" | "#"               { commentaire_Ligne lexbuf }
  | "/*"                     { commentaire_C lexbuf }
  | "(*"                     { commentaire_Caml lexbuf }
  | "(#"                     { commentaire_Imoview lexbuf }

  | eof                      { TokFin }



and strparse = parse
    '"'
      { () }

  | '\\' ("\010" | "\013" | "\013\010") [' ' '\009'] *
      { strparse lexbuf }
  | '\\' ['\\' '"' 'n' 't' 'b' 'r']
      { store_string_char(char_for_backslash(Lexing.lexeme_char lexbuf 1));
        strparse lexbuf }
  | '\\' ['0'-'9'] ['0'-'9'] ['0'-'9']
      { store_string_char(char_for_decimal_code lexbuf 1);
         strparse lexbuf }
  | eof
      { raise (Erreur "Unterminated string")}
  | _
      { store_string_char(Lexing.lexeme_char lexbuf 0);
        strparse lexbuf }

(**************************)
(* Lexers de commentaires *)
(**************************)


(* Commentaires en /* ... */ *)
and commentaire_C = parse
    "*/"                     { token lexbuf }
  | [^'*' '\n']+             { commentaire_C lexbuf }
  | ['*']+[^'*' '/' '\n']+   { commentaire_C lexbuf }
  | ['*']+ '\n'              { incr Parametres.line; commentaire_C lexbuf }
  | '\n'                     { incr Parametres.line; commentaire_C lexbuf }
  | eof                      { TokFin }



(* Commentaires en (* ... *) *)
and commentaire_Caml = parse
    "*)"                     { token lexbuf }
  | [^'*' '\n']+             { commentaire_Caml lexbuf }
  | ['*']+[^'*' ')' '\n']+   { commentaire_Caml lexbuf }
  | ['*']+ '\n'              { incr Parametres.line; commentaire_Caml lexbuf }
  | '\n'                     { incr Parametres.line; commentaire_Caml lexbuf }
  | eof                      { TokFin }



(* Commentaires en (# ... #) *)
and commentaire_Imoview = parse
    "#)"                     { token lexbuf }
  | [^'#' '\n']+             { commentaire_Imoview lexbuf }
  | ['#']+[^'#' ')' '\n']+   { commentaire_Imoview lexbuf }
  | ['#']+ '\n'              { incr Parametres.line; commentaire_Imoview lexbuf }
  | '\n'                     { incr Parametres.line; commentaire_Imoview lexbuf }
  | eof                      { TokFin }



(* Commentaires en # ... \n ou // ... \n *)
and commentaire_Ligne = parse
    [^'\n']                  { commentaire_Ligne lexbuf }
  | '\n'                     { incr Parametres.line; token lexbuf }
  | eof                      { TokFin }
