(*************************************************************************)
(*                                                                       *)
(* Ce fichier contient les fonction de gestion de la file des commandes, *)
(* que ce soit pour ajouter une commande ou pour les traiter.            *)
(*                                                                       *)
(*************************************************************************)



open Types
open File
open Triplet



let animToString etat sens = !Parametres.ecrireStatut (match etat, sens
						       with
							 | Lct   , Avant
							 | Step  , Avant   -> "Lecture"
							 | Lct   , Arriere
							 | Step  , Arriere -> "Lecture arrire"
							 | Rapide, Avant   -> "Avance rapide"
							 | Rapide, Arriere -> "Retour rapide"
							 | Stop  , _       -> "Stop"
							 | Tempo , _       -> "Pause"
							 | Fin   , _       -> "Fin"
							 | _               -> "Erreur"
						      )



let doStop () = while not (Parametres.command_done#isEmpty)
		do
		  Parametres.command_todo#pushBegin (Parametres.command_done#popBegin ())
		done;
		List.iter (fun scn -> scn#hide) !Parametres.dessin



(*****************************************************)
(* Fonction  timer  qui  ralise  le  mouvement  mvt *)
(* jusqu' son terme.                                *)
(* oldTime   est  l'heure    laquelle   le  dernier *)
(* raffraichissement du mouvement a t effectu.    *)
(* next  est  l'argument    passer    la  fonction *)
(* sceneTimer une fois le mouvement termin.         *)
(*****************************************************)
(* Type : mouvement -> float -> anim -> unit -> unit *)
(*****************************************************)
let rec move_func mvt oldTime next () = (* On supprime le timout pour viter les saturations *)
					GMain.Timeout.remove (unSome !Parametres.timer);
					Parametres.timer := None;

					let amplitude vitesse duree limite = min_triplet (mult_triplet (vitesse)#getAll duree) (limite)#getAll
					in
					  (* Si le mouvement n'est pas fini, on continue *)
					  if !Parametres.state <> Mvt ||
					     (((mvt#transDelta)#getAll = (0.0, 0.0, 0.0)) &&
					      ((mvt#  rotDelta)#getAll = (0.0, 0.0, 0.0)) &&
					      ((mvt# spinDelta)#getAll = (0.0, 0.0, 0.0)))
					  then
					    begin
					      checkNewState next
					    end
					  else
					    begin
					      let actuTime = Unix.gettimeofday ()
					      in
					      (* Calcul de la dure depuis le dernier raffraichissement pour *)
					      (* connatre l'amplitude du morceau de mouvement  raliser.   *)
					      let deltaT = actuTime -. oldTime
					      in
					      (* Calcule des amplitudes de mouvement  raliser pour chaque type de mouvement *)
					      let trans = amplitude mvt#transSpeed deltaT mvt#transDelta
					      and rot   = amplitude mvt#  rotSpeed deltaT mvt#  rotDelta
					      and spin  = amplitude mvt# spinSpeed deltaT mvt# spinDelta
					      in
						(* Mise  jour des amplitudes restant  raliser *)
						(mvt#transDelta)#setAll (diff_triplet (mvt#transDelta)#getAll trans);
						(mvt#  rotDelta)#setAll (diff_triplet (mvt#  rotDelta)#getAll rot  );
						(mvt# spinDelta)#setAll (diff_triplet (mvt# spinDelta)#getAll spin );
						(* Ralisation des mouvements *)
						if !Parametres.sens = Avant
						then
						  begin
						    Move.translation trans;
						    Move.rotation    rot;
						    Move.spin        spin (0.0, 0.0, 0.0)
						  end
						else
						  begin
						    Move.translation (mult_triplet trans (-1.0));
						    Move.rotation    (mult_triplet rot   (-1.0));
						    Move.spin        (mult_triplet spin  (-1.0)) (0.0, 0.0, 0.0)
						  end;
						if next <> Rapide
						then
						  Draw.draw_scene ();

						Parametres.timer := Some (GMain.Timeout.add ~ms:1
											    ~callback:(move_func mvt actuTime next)
									 )
					    end;
					  true



and checkNewState etat = match !Parametres.nextState
			 with
			   | None            -> (match etat
						 with
						   | Lct   -> Parametres.timer := Some (GMain.Timeout.add ~ms:!Parametres.cmdTimout
													  ~callback:(sceneTimer Lct)
										       )

						   | Fin
						   | Rapide
						   | Step
						   | Stop
						   | Tempo -> animToString etat !Parametres.sens;
							      Parametres.state := etat

						   | Mvt   -> ()
						)

			   | Some (nxt, dir) -> animToString nxt dir;
						Parametres.nextState := None;
						Parametres.sens      := dir;
						Parametres.timer     := Some (GMain.Timeout.add ~ms:10
												~callback:(sceneTimer nxt)
									     )



and exec_command cmd etat = let masque dir lst = List.iter (match dir
							    with
							      | Avant   -> (fun scn -> scn#hide)
							      | Arriere -> (fun scn -> scn#show)
							   )
							   lst
			    and sensOppose dir = match dir
						 with
						   | Avant   -> Arriere
						   | Arriere -> Avant
			    in
			      match cmd
			      with
				| Pause                     -> Tempo

				| Show scnLst               -> masque (sensOppose !Parametres.sens) scnLst;
							       etat

				| Hide scnLst               -> masque             !Parametres.sens  scnLst;
							       etat

				| Replace (hide, show)      -> masque (sensOppose !Parametres.sens) show;
							       masque             !Parametres.sens  (unSome hide);
							       etat

				| Command str               -> if !Parametres.sens = Avant
							       then
								 begin
								   print_endline str;
								   flush stdout;
								   ignore (Sys.command str)
								 end;
							       etat

				| View (viewr, centr, vert) -> (match !Parametres.sens
								with
								  | Avant   -> (* On fixe le centre de la vision *)
									       Parametres.centre#setAll (match centr
													 with
													   | AutoCenter -> Parametres.milieu_dessin#getAll
													   | Center pt  -> pt#getAll
													);

									       (* On fixe le vecteur Z du repre de la camera *)
									       (match viewr
										with
										  | Viewer pt  -> let vecteur = diff_triplet pt#getAll Parametres.centre#getAll
												  in
												  let norme = normale_triplet vecteur
												  in
												    Parametres.zoom := norme;
												    Parametres.vectZ#setAll (div_triplet vecteur norme)
										  | AutoViewer -> Parametres.zoom := !Parametres.rayon;
												  Parametres.vectZ#setAll (0.0, 0.5, 0.86602540378445)
									       );

									       (* On fixe la verticale du repre de la camera *)
									       let z = Parametres.vectZ#getAll
									       in
									       let vect = (match vert
											   with
											     | AutoVertical -> if (colineaire (0.0, 1.0, 0.0) z)
													       then (0.0, 0.0, -1.0)
													       else (0.0, 1.0,  0.0)
											     | Vertical pt  -> (if (colineaire pt#getAll z)
														then if (pt#getAll = (0.0, 1.0, 0.0))
														     then (0.0, 0.0, -1.0)
														     else (0.0, 1.0,  0.0)
														else pt#getAll
													       )
											  )
									       in
									       let sbt = normalise_triplet (diff_triplet vect (mult_triplet z (scalaire vect z)))
									       in
										 Parametres.vectY#setAll sbt;

										 (* On fixe le verteur X du repre de la camera *)
										 Parametres.vectX#setAll (prod_triplet sbt z)

								  | Arriere -> ()
							       );
							       etat

				| Movement mvt              -> (* On ajoute un timout ralisant le mouvement *)
							       let copie = new Types.mouvement
							       in
								 copie#setrotSpeed   (Oo.copy mvt#rotSpeed);
								 copie#settransSpeed (Oo.copy mvt#transSpeed);
								 copie#setspinSpeed  (Oo.copy mvt#spinSpeed);
								 copie#setrotDelta   (Oo.copy mvt#rotDelta);
								 copie#settransDelta (Oo.copy mvt#transDelta);
								 copie#setspinDelta  (Oo.copy mvt#spinDelta);
								 Parametres.timer := Some (GMain.Timeout.add ~ms:1
													     ~callback:(move_func copie (Unix.gettimeofday ()) etat)
											  );
								 Mvt



and sceneTimer etat () = (* On supprime le timout pour viter les saturations *)
			 GMain.Timeout.remove (unSome !Parametres.timer);
			 Parametres.timer := None;

			 let (cmdTodo, cmdDone) = (match !Parametres.sens
						   with
						     | Avant   -> Parametres.command_todo, Parametres.command_done
						     | Arriere -> Parametres.command_done, Parametres.command_todo
						  )
			 in
			   checkNewState (match etat
					  with
					    | Stop   -> doStop ();
							Draw.draw_scene ();
							Stop

					    | Rapide -> while not (cmdTodo#isEmpty)
							do
							  let cmd = cmdTodo#popBegin ()
							  in
							    ignore (exec_command cmd Rapide);
							    cmdDone#pushBegin cmd;
							done;
							Draw.draw_scene ();
							Tempo

					    | Step   -> if not (cmdTodo#isEmpty)
							then
							  begin
							    let cmd = cmdTodo#popBegin ()
							    in
							      cmdDone#pushBegin cmd;
							      let nxt = exec_command cmd Tempo
							      in
								Draw.draw_scene ();
								nxt
							  end
							else
							  if !Parametres.sens = Avant
							  then
							    Fin
							  else
							    Tempo

					    | Lct    -> if not (cmdTodo#isEmpty)
							then
							  begin
							    let cmd = cmdTodo#popBegin ()
							    in
							      cmdDone#pushBegin cmd;
							      let nxt = exec_command cmd Lct
							      in
								Draw.draw_scene ();
								nxt
							  end
							else
							  if !Parametres.sens = Avant
							  then
							    Fin
							  else
							    Tempo

					    | Tempo  -> Tempo

					    | _      -> (* Ne doit pas arriver :$ *)
							Tempo
					 );
			   true



(***********************************************************)
(* Calcule le centre d'un BoundingBox                      *)
(***********************************************************)
(* Type : float * float * float * float * float * float -> *)
(*        float * float * float                            *)
(***********************************************************)
let milieu_bounding (minX, minY, minZ, maxX, maxY, maxZ) =
  let x = 0.5 *. (maxX +. minX)
  and y = 0.5 *. (maxY +. minY)
  and z = 0.5 *. (maxZ +. minZ)
  in
    (x, y, z)



(***************************************************************)
(* Calcule le rayon de la sphre circonscrite d'un BoundingBox *)
(***************************************************************)
(* Type : float * float * float * float * float * float ->     *)
(*        float                                                *)
(***************************************************************)
let rayon_bounding (minX, minY, minZ, maxX, maxY, maxZ) =
  let x = 0.5 *. (maxX -. minX)
  and y = 0.5 *. (maxY -. minY)
  and z = 0.5 *. (maxZ -. minZ)
  in
    sqrt (x*.x +. y*.y +. z*.z)



(*************************************************************)
(* Ajoute la commande cmd  la file d'attente des commandes. *)
(* Si la file est vide et qu'aucune pause n'est active, on   *)
(* lance le gestionnaire de commandes.                       *)
(*************************************************************)
(* Type : commande -> unit                                   *)
(*************************************************************)
let add_command cmd = (match cmd
		       with
			 | Hide scnLst          -> let (newHideLst, newShownLst) = List.partition (fun scn -> List.exists (fun s -> s=scn) scnLst) !Parametres.lstShownEnd
						   in
						     if newHideLst <> []
						     then
						       begin
							 Parametres.lstShownEnd := newShownLst;
							 Parametres.command_todo#pushEnd (Hide newHideLst)
						       end

			 | Show scnLst          -> let newShowLst = List.filter (fun scn -> not (List.exists (fun s -> s=scn) !Parametres.lstShownEnd)) scnLst
						   in
						     if newShowLst <> []
						     then
						       begin
							 Parametres.lstShownEnd := newShowLst @ !Parametres.lstShownEnd;
							 Parametres.command_todo#pushEnd (Show newShowLst)
						       end

			 | Replace (hide, show) -> let lstHide = List.filter (fun scn -> not (List.exists (fun s -> s=scn) show)) (match hide
																   with
																     | None   -> !Parametres.dessin
																     | Some h -> h
																  )
						   in
						   let (newHideLst, tmp) = List.partition (fun scn -> List.exists (fun s -> s=scn) lstHide) !Parametres.lstShownEnd
						   in
						   let newShowLst = List.filter (fun scn -> not (List.exists (fun s -> s=scn) tmp)) show
						   in
						     Parametres.lstShownEnd := newShowLst @ tmp;
						     (match (newHideLst, newShowLst)
						      with
							| [], [] -> ()
							| [],  _ -> Parametres.command_todo#pushEnd (Show newShowLst)
							|  _, [] -> Parametres.command_todo#pushEnd (Hide newHideLst)
							|  _,  _ -> Parametres.command_todo#pushEnd (Replace (Some newHideLst, newShowLst))
						     )

			 | _                    -> Parametres.command_todo#pushEnd cmd
		      );
		      (*if Parametres.command_todo#length = 1 && !Parametres.state = Fin
		      then !(Parametres.lecture)*) ()



(************************************************)
(* Ajoute la scne scn  la liste des scnes et *)
(* ajoute la  commande qui affiche la scne.    *)
(* Si la file d'attente des commandes est vide  *)
(* et qu'aucune pause n'tait active, on lance  *)
(* le gestionnaire de commandes.                *)
(************************************************)
(* Type : commande -> unit                      *)
(************************************************)
let add_scene scn =
  Parametres.dessin := scn::!Parametres.dessin;
  let bb = Boundingbox.boundingScene scn
  in
    (* Si il s'agit de la 1re scne, on met  jour le rayon  et le zoom *)
    if (List.length !Parametres.dessin) = 1
    then
      begin
	let (x, y, z) = milieu_bounding bb
	in
	  Parametres.milieu_dessin#setAll (x, y, z);
	  Parametres.rayon := max 200.0 (4.0 *. (rayon_bounding bb) +. (sqrt (x*.x +. y*.y +. z*.z)));
	  Parametres.zoom  := !Parametres.rayon
      end;
