(************************************************)
(*                                              *)
(* Ce fichier contient des fonction utiles pour *)
(* travailler sur des triplets de floattants.   *)
(*                                              *)
(************************************************)



(****************************************************)
(* Multiplie les coordonnes d'un triplet (x, y, z) *)
(* par un coefficiant c.                            *)
(****************************************************)
(* Type : float * float * float -> float ->         *)
(*        float * float * float                     *)
(****************************************************)
let mult_triplet (x, y, z) c = (c*.x, c*.y, c*.z)



(*************************************************)
(* Divise les coordonnes d'un triplet (x, y, z) *)
(* par un coefficiant c.                         *)
(*************************************************)
(* Type : float * float * float -> float ->      *)
(*        float * float * float                  *)
(*************************************************)
let div_triplet (x, y, z) c = (x/.c, y/.c, z/.c)



(************************************************************)
(* Additionne deux triplets (x1, y1, z1) et (x2, y2, z2).   *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        float * float * float                             *)
(************************************************************)
let add_triplet (x1, y1, z1) (x2, y2, z2) = (x1+.x2, y1+.y2, z1+.z2)



(************************************************************)
(* Soustrait deux triplets (x1, y1, z1) et (x2, y2, z2).    *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        float * float * float                             *)
(************************************************************)
let diff_triplet (x1, y1, z1) (x2, y2, z2) = (x1-.x2, y1-.y2, z1-.z2)



(*********************************************************)
(* Retourne l'oppos d'un triplet.                       *)
(*********************************************************)
(* Type : float * float * float -> float * float * float *)
(*********************************************************)
let oppose_triplet (x, y, z) = (-.x, -.y, -.z)



(*******************************************************)
(* Retourne la 1re coordonne d'un triplet (x, y, z). *)
(*******************************************************)
(* Type : float * float * float -> float               *)
(*******************************************************)
let abscisse (x, _, _) = x



(*******************************************************)
(* Retourne la 2me coordonne d'un triplet (x, y, z). *)
(*******************************************************)
(* Type : float * float * float -> float               *)
(*******************************************************)
let ordonnee (_, y, _) = y



(*******************************************************)
(* Retourne la 3me coordonne d'un triplet (x, y, z). *)
(*******************************************************)
(* Type : float * float * float -> float               *)
(*******************************************************)
let cote (_, _, z) = z



(********************************************************************)
(* Calcule le carr de la norme euclidienne d'un vecteur (x, y, z). *)
(********************************************************************)
(* Type : float * float * float -> float                            *)
(********************************************************************)
let normale_carre_triplet (x, y, z) = x*.x +. y*.y +. z*.z



(********************************************************)
(* Calcule la norme euclidienne d'un vecteur (x, y, z). *)
(********************************************************)
(* Type : float * float * float -> float                *)
(********************************************************)
let normale_triplet (x, y, z) = sqrt (x*.x +. y*.y +. z*.z)



(***********************************************************)
(* Calcule le vecteur unitaire de mme sens que (x, y, z). *)
(***********************************************************)
(* Type : float * float * float -> float * float * float   *)
(***********************************************************)
let normalise_triplet t = div_triplet t (normale_triplet t)



(************************************************************)
(* Calcule le produit scalaire de 2 vecteurs (x1, y1, z1)   *)
(* et (x2, y2, z2).                                         *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        float                                             *)
(************************************************************)
let scalaire (x1, y1, z1) (x2, y2, z2) = x1*.x2 +. y1*.y2 +. z1*.z2



(************************************************************)
(* Calcule le produit vectoriel de 2 vecteurs (x1, y1, z1)  *)
(* et (x2, y2, z2).                                         *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        float * float * float                             *)
(************************************************************)
let prod_triplet (x1, y1, z1) (x2, y2, z2) =
  (y1*.z2 -. z1*.y2,
   z1*.x2 -. x1*.z2,
   x1*.y2 -. y1*.x2)



(************************************************************)
(* Teste si 2 vecteures (x1, y1, z1) et (x2, y2, z2) sont   *)
(* colinaires.                                             *)
(************************************************************)
(* Type : float * float * float -> float * float * float -> *)
(*        bool                                              *)
(************************************************************)
let colineaire t1 t2 = ((prod_triplet t1 t2) == (0.0, 0.0, 0.0))



(****************************************************************)
(* Calcule le triplet compos pour chaque coordonne de la plus *)
(* petites valeurs de 2 vecteurs (x1, y1, z1) et (x2, y2, z2).  *)
(****************************************************************)
(* Type : float * float * float -> float * float * float ->     *)
(*        float * float * float                                 *)
(****************************************************************)
let min_triplet (x1, y1, z1) (x2, y2, z2) = (min x1 x2, min y1 y2, min z1 z2)



(*****************************************)
(* Convertit un triplet en point.        *)
(*****************************************)
(* Type : float * float * float -> point *)
(*****************************************)
let triplet2point (x, y, z) = new Types.point x y z
