(*******************************************************************)
(*                                                                 *)
(* Ce fichier contient l'ensemble des types utiliss dans Imoview. *)
(*                                                                 *)
(*******************************************************************)



(**************************************)
(* Types de fichiers images supports *)
(**************************************)
type format = 
   | Pdf
   | PsVect
   | EpsVect
   | Bmp
   | Gif
   | Jpeg
   | Ppm
   | Png
   | Ps



(***********************)
(* Sens de l'animation *)
(***********************)
type sens =
  | Avant
  | Arriere



(***********************)
(* Etat de l'animation *)
(***********************)
type anim =
  | Fin    (* Arrt car la pile des commandes est vide *)
  | Lct    (* Lecture -> excution d'un autre type de commande *)
  | Rapide (* Avance ou Retour rapide *)
  | Stop   (* Arrt de l'animation *)
  | Tempo  (* Pause -> attente d'un appui sur ESPACE *)
  | Step   (* N'effectue qu'une seule commande *)
  | Mvt    (* L'animation est en cour (le thread est dj lanc) *)



(*********************************)
(* Etat d'un bouton de la souris *)
(*********************************)
type mouseState =
  | DOWN
  | UP



(*****************************)
(* Type d'entre des donnes *)
(*****************************)
type input =
  | File
  | Stdin



(*********************************************************)
(* Type d'affichage des sphres, botes et ellipsodes : *)
(*  - All    -> entier                                   *)
(*  - Inf    -> moitie infrieure                       *)
(*  - Sup    -> moitie suprieure                       *)
(*  - Bottom -> Fond                                     *)
(*********************************************************)
type cont_mode =
  | All
  | Inf
  | Sup
  | Bottom



(**************************************)
(* Etat de chaque bouton de la souris *)
(**************************************)
class mouseButton =
object
  (* Position x en coordonne de fentre lors du clic *)
  val mutable x     = 0.0
  (* Position y en coordonne de fentre lors du clic *)
  val mutable y     = 0.0
  (* Etat du clic : DOWN ou UP *)
  val mutable state = UP
  (* Dtermine si la touche SHIFT tait enfonce lors du du clic *)
  val mutable shift = false
  (* Dtermine si la touche CTRL tait enfonce lors du du clic *)
  val mutable ctrl  = false

  (* Mthodes pour accder et modifier les donnes *)
  method getX              = x
  method getY              = y
  method setX newX         = x <- newX
  method setY newY         = y <- newY
  method getState          = state
  method setState newState = state <- newState
  method getShift          = shift
  method setShift newShift = shift <- newShift
  method getCtrl           = shift
  method setCtrl   newCtrl = ctrl  <- newCtrl
end



(*********)
(* Point *)
(*********)
class point (initX:float) (initY:float) (initZ:float) =
object
  (* Coordonne suivant X *)
  val mutable x = initX
  (* Coordonne suivant Y *)
  val mutable y = initY
  (* Coordonne suivant Z *)
  val mutable z = initZ

  (* Mthodes pour accder et modifier les coordonnes une par une *)
  method getX                   = x
  method getY                   = y
  method getZ                   = z
  method setX newX              = x <- newX
  method setY newY              = y <- newY
  method setZ newZ              = z <- newZ

  (* Mthodes pour accder et modifier toutes les coordonnes *)
  method getAll                 = (x, y, z)
  method setAll (xx, yy, zz)    = x <- xx;
				  y <- yy;
				  z <- zz

  (* Mthodes pour incrmenter les coordonnes *)
  method updateX n              = x <- x +. n
  method updateY n              = y <- y +. n
  method updateZ n              = z <- z +. n
  method updateAll (xx, yy, zz) = x <- x +. xx;
				  y <- y +. yy;
				  z <- z +. zz
  (* Mthodes opposes *)
  method oppose () =    x<-(-.x);
                        y<-(-.y);
                        z<-(-.z)
                          
  (* Dtermine  il s'agit d'une valeur par dfaut (non fixe) *)
  method isDefault = (x == nan || x == nan || x == nan)
end



(****************************************************************)
(* Couleur                                                      *)
(* Une couleur par dfaut comporte une composante nan.          *)
(* Une couleur valide a ses compossantes dans l'intervale [0;1] *)
(****************************************************************)
class couleur (initR:float) (initV:float) (initB:float) (initAlpha:float) =
object
  (* Composante rouge *)
  val mutable rouge = initR
  (* Composante verte *)
  val mutable vert  = initV
  (* Composante bleu *)
  val mutable bleu  = initB
  (* Transparence *)
  val mutable alpha = initAlpha

  (* Mthodes pour accder et modifier les composantes de la couleur *)
  method getRouge                = rouge
  method getVert                 = vert
  method getBleu                 = bleu
  method getAlpha                = alpha
  method getAll                  = (rouge, vert, bleu, alpha)
  method setRouge newRouge       = rouge <- newRouge
  method setVert newVert         = vert <- newVert
  method setBleu newBleu         = bleu <- newBleu
  method setAlpha newAlpha       = alpha <- newAlpha
  method setAll (rg, vt, bl, al) = rouge <- rg;
				   vert  <- vt;
				   bleu  <- bl;
				   alpha <- al

  (* Dtermine  il s'agit d'une couleur par dfaut (non fixe) *)
  method isDefault               = (rouge == nan || vert == nan || bleu == nan || alpha == nan)
end



(**************)
(* Primitives *)
(**************)
type primitive =
  | Box      of point                                                  (* taille *)
  | Cone     of float ref * float ref * bool ref * int ref             (* rayon  hauteur  solide  slice *)
  | Cylinder of float ref * float ref * bool ref * int ref             (* rayon  hauteur  solide  slice *)
  | Disc     of float ref * int ref                                    (* rayon  slice *)
  | Empty    of point * bool ref                                       (* dimension  visible *)
  | Frustum  of float ref * float ref * float ref * bool ref * int ref (* rayon  hauteur  rate  solide  slice *)
  | Polygone of point list ref * point                                 (* pointlist  normale *)
  | Polyline of point list ref                                         (* pointlist *)
  | Sphere   of float ref * int ref * int ref                          (* rayon  slice  stack *)
  | String   of (([`rgba], [`float]) GlPix.t) ref                       (* le pixmap du texte (calcul dans parser.mly *)


(********)
(* Axes *)
(********)
type axe =
  | AxeX
  | AxeY
  | AxeZ
  | AxeNone (* AxeNone sert juste dans le parser de valeur par dfaut *)



(*******************)
(* Transformations *)
(*******************)
type transformation =
  | Rotation    of axe * float (* axe  angle *)
  | Scale       of point       (* taux *)
  | Translation of point       (* vecteur *)



(**************)
(* Alignement *)
(**************)
type align =
  | Left
  | Middle
  | Right



(*********)
(* Grids *)
(*********)
type grid =
    (* alignement1  alignement2  alignement3 *)
  | Grid0 of align * align * align
    (* alignement1  alignement2  space1 *)
  | Grid1 of axe * align * align * float
    (* axis1  axis2  alignement1  alignement2  alignement3  space1  space2  ratio1  ratio2 *)
  | Grid2 of axe * axe * align * align * align * float * float * int * int
    (* axis1  axis2  alignement1  alignement2  alignement3  space1  space2  space3  ratio1  ratio2  ratio3 *)
  | Grid3 of axe * axe * align * align * align * float * float * float * int * int * int
    (* axis1  axis2  alignement3  space  ratio1  ratio2 *)
  | Hexa2 of axe * axe * align * float * int * int
    (* axis1  axis2  space  ratio1  ratio2  ratio3 *)
  | Hexa3 of axe * axe * float * int * int * int



(************************)
(* Forme des Containers *)
(************************)
type typeContainer =
  | ContSphere of float ref
  | ContEllipsoide of point
  | ContBox of point



(***************************************)
(* Attribut d'affichage des Containers *)
(***************************************)
type draw =
  | Blend of float
  | Mixed of draw * draw (* Inf Sup *)
  | Shadow
  | Open
  | Wire
  | Close



(*********)
(* Objet *)
(*********)
type objet =
  | Primitive      of primitive * couleur * draw ref * ((point * float ref * objet list) option)  (* primitive  couleur  fil-de-fer  contenu *)
  | Transformation of transformation * objet
  | Groupe         of couleur * objet list                                                        (* couleur  coin1  coin2  liste *)
  | Grid           of point * grid * couleur * (objet * point) list                               (* origine  paramtres  couleur  coin1  coin2  gridlist *)
  | Vide                                                                                          (* objet vide :-| *)



(*********)
(* Scne *)
(*********)
class scene =
object
  (* Dtermine si la scne est affiche ou non *)
  val mutable shown   = false
  (* Liste des objets de la scne *)
  val mutable contenu = ([]:(objet list))
  (* Nom de la scne               *)
  (* "" si la scne n'a pas de nom *)
  val mutable name    = ""

  (* Mthodes pour accder et modifier les attributs de la scne *)
  method show                  = shown <- true
  method hide                  = shown <- false
  method isShown               = shown
  method getName               = name
  method setName newName       = name <- newName
  method getContenu            = contenu
  method setContenu newContenu = contenu <- newContenu
end



(********************)
(* Vue de la camera *)
(********************)
type viewer =
    (* Vue automatique *)
  | AutoViewer
    (* Vue personnalise *)
  | Viewer of point



(********************)
(* Centre de la vue *)
(********************)
type center =
    (* Centre automatique *)
  | AutoCenter
    (* Centre personnalis *)
  | Center of point



(***********************)
(* Verticale de la vue *)
(***********************)
type vertical =
    (* Verticale automatique *)
  | AutoVertical
    (* Verticale personnalis *)
  | Vertical of point



(**************************)
(* Mouvement de la camera *)
(**************************)
class mouvement =
object
  (* Vitesse de translation *)
  val mutable _transSpeed = new point 0.0 0.0 0.0
  (* Amplitude de translation *)
  val mutable _transDelta = new point 0.0 0.0 0.0
  (* Vitesse de rotation par rapport au centre de la vue *)
  val mutable   _rotSpeed = new point 0.0 0.0 0.0
  (* Amplitude de rotation par rapport au centre de la vue *)
  val mutable   _rotDelta = new point 0.0 0.0 0.0
  (* Vitesse de rotation de la scne sur elle-mme *)
  val mutable  _spinSpeed = new point 0.0 0.0 0.0
  (* Amplitude de rotation de la scne sur elle-mme *)
  val mutable  _spinDelta = new point 0.0 0.0 0.0

  (* Mthodes pour accder aux attributs du mouvement *)
  method transSpeed = _transSpeed
  method   rotSpeed =   _rotSpeed
  method  spinSpeed =  _spinSpeed
  method transDelta = _transDelta
  method   rotDelta =   _rotDelta
  method  spinDelta =  _spinDelta

  method inverse()  =   
                        _transDelta#oppose();
                        _rotDelta#oppose();
                        _spinDelta#oppose();
                        _transSpeed#oppose();
                        _rotSpeed#oppose();
                        _spinSpeed#oppose()
  
  method settransSpeed x = _transSpeed <- x
  method setrotSpeed   x = _rotSpeed   <- x
  method setspinSpeed  x = _spinSpeed  <- x
  method settransDelta x = _transDelta <- x
  method setrotDelta   x = _rotDelta   <- x
  method setspinDelta  x = _spinDelta  <- x
  
end



(*************)
(* Commandes *)
(*************)
type commande =
    (* On affiche les scnes contenues dans la liste *)
  | Show     of scene list
    (* On cache les scnes contenues dans la liste *)
  | Hide     of scene list
    (* On substitue l'affichage de scnes par d'autre.     *)
    (* None -> toutes les scnes affiches sont remplaces *)
  | Replace  of scene list option * scene list
    (* Modification de la vue *)
  | View     of viewer * center * vertical
    (* On lance une commande shell *)
  | Command  of string
    (* On effectue un mouvement *)
  | Movement of mouvement
    (* Pause jusqu' ce que l'on appuie sur ESPACE *)
  | Pause



(******************)
(* Les exceptions *)
(******************)
exception ParseError of string
exception Erreur     of string
exception Help
exception APas

let unSome x = match x
	       with
		 | None   -> raise APas
		 | Some y -> y
